/*
 ****************************************************************************
 * Fits Data View
 *
 * Input:   Image files
 *
 * Output:  A list with image keywords and statistics
 *
 * Copyright (C) 2018, Hartmut V. Bornemann
 *
 * Joined development with Herbert Walter
 *
 * Version history
 *          1.0.7    2021-07-16 variable csv-separators added
 *          1.0.6    2021-05-18 auto read recent files removed
 *          1.0.5    2020-02-27 section name set to 'Fits Data View'
 *          1.0.4    2018-10-07 formatting filenames in full or partial version
 *                              progress number reset with filetabel changes
 *                              trimmed single quotes from FITS values
 *          1.0.3    2018-10-02 added table output precision and column sorting
 *          1.0      2018-07-25 first release
 *
 ****************************************************************************
 */

#feature-id Image Analysis > Fits Data View

#include <pjsr/FileMode.jsh>
#include <pjsr/Sizer.jsh>
#include <pjsr/DataType.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/StdButton.jsh>
#include <pjsr/StdIcon.jsh>
#include <pjsr/StdCursor.jsh>
#include <pjsr/UndoFlag.jsh>
#include <pjsr/StarDetector.jsh>
#include <pjsr/NumericControl.jsh>

#define USE_STAR_DETECTOR true
#define __PJSR_NO_STAR_DETECTOR_TEST_ROUTINES

#define TITLE "Fits Data View"
#define VERSION "1.0.7"
#feature-info Fits Data View.<br>\
   <br>\Written by Hartmut V. Bornemann.<br>\
   <br>\
   References: None<br>\
   <br>\
#define strict 0
#define SETKEY "FDVDialog"
// Shadows clipping point in (normalized) MAD units from the median.
#define DEFAULT_AUTOSTRETCH_SCLIP  -2.80
// Target mean background in the [0,1] range.
#define DEFAULT_AUTOSTRETCH_TBGND   0.25
#define DEFAULT_AUTOSTRETCH_CLINK   true


// ***************************************************************************
//
// Fits Keyword Cross Reference
//
// ***************************************************************************

// ***************************************************************************
// Statistical functions
// ***************************************************************************

function avgDev(fo)
{
	return fo.statistics.avgDev * fo.dialog.numTrans;
}

function BWMV(fo)
{
	return  Math.sqrt(fo.statistics.bwmv) * fo.dialog.numTrans;
}

function count(fo)
{
	return fo.statistics.count;
}

function eccentricities(fo)
{
	var e = fo.Eccentricities();
	return e[0];
}

function fwhm_(fo)
{
	var e = fo.FWHM();
	return e[0];
}

function MAD(fo)
{
	//return fo.statistics.mad;
	return fo.image.MAD() * fo.dialog.numTrans;
}

function maximum(fo)
{
	return fo.statistics.maximum * fo.dialog.numTrans;
}

function maximumPosition(fo)
{
	var p = fo.statistics.maximumPosition;
	return p.x + ", " + p.y;
}

function mean(fo)
{
	return fo.statistics.mean * fo.dialog.numTrans;
}

function meanOfSquares(fo)
{
	return fo.image.meanOfSquares() * fo.dialog.numTrans;
}

function median(fo)
{
	return fo.statistics.median * fo.dialog.numTrans;
}

function minimum(fo)
{
	return fo.statistics.minimum * fo.dialog.numTrans;
}

function minimumPosition(fo)
{
	var p = fo.statistics.minimumPosition;
	return p.x + ";" + p.y;
}

function modulus(fo)
{
	return fo.image.modulus() * fo.dialog.numTrans;
}

function norm(fo)
{
	return fo.image.norm() * fo.dialog.numTrans;
}

function PBMV(fo)
{
	return  Math.sqrt(fo.statistics.pbmv) * fo.dialog.numTrans;
}

function Qn(fo)
{
	return fo.image.Qn() * fo.dialog.numTrans;
}

function residuals(fo)
{
	var e = fo.Residuals();
	return e[0];
}

function Sn(fo)
{
	return fo.image.Sn() * fo.dialog.numTrans;
}

function snr(fo)
{
	return SNREvaluation (fo.image);
}

function stdDev(fo)
{
	return fo.statistics.stdDev * fo.dialog.numTrans;
}

function sumOfSquares(fo)
{
	return fo.statistics.sumOfSquares * fo.dialog.numTrans;
}

function variance(fo)
{
	return fo.statistics.variance * fo.dialog.numTrans * fo.dialog.numTrans;
}

function background(fo)
{
	var b = 3 * fo.statistics.median - 2 * fo.statistics.mean;
	return b * fo.dialog.numTrans;
}
// ***************************************************************************

function fillFunctions(tree)
{
	//
	// addFunction(tree, description, functionName)
	//
	addFunction(tree, "Average deviation (avgDev)", "avgDev");
	addFunction(tree, "Biweight midvariance sqrt(BWMV)", "BWMV");
	addFunction(tree, "Count [pixels]", "count");
	addFunction(tree, "Excentricity", "eccentricities");
	addFunction(tree, "FWHM [pixel]", "fwhm_");
	addFunction(tree, "Image Backgound Mean", "background");
	addFunction(tree, "Maximum", "maximum");
	addFunction(tree, "MaximumPosition", "maximumPosition");
	addFunction(tree, "Mean", "mean");
	addFunction(tree, "MeanOfSquares", "meanOfSquares");
	addFunction(tree, "Median absolute deviation (MAD)", "MAD");
	addFunction(tree, "Median", "median");
	addFunction(tree, "Minimum", "minimum");
	addFunction(tree, "MinimumPosition", "minimumPosition");
	addFunction(tree, "Modulus", "modulus");
	addFunction(tree, "Norm", "norm");
	addFunction(tree, "Percentage bend midvariance sqrt(PBMV)", "PBMV");
	addFunction(tree, "Qn", "Qn");
	addFunction(tree, "Residuals", "residuals");
	addFunction(tree, "Sn", "Sn");
	addFunction(tree, "Signal to noise ratio [decibels]", "snr");
	addFunction(tree, "Standard deviation (stdDev)", "stdDev");
	addFunction(tree, "Sum of squares", "sumOfSquares");
	addFunction(tree, "Variance", "variance");
}

function evaluateFunction(name, v)
{
	return this[name](v);
}


// ***************************************************************************

function tableGenerator(dialog)
{
   //
   // create an Array with datatypes
   //
	this.table  = [];			// 1 row per file, this is the table body
   this.header = [];       // text
	var keyNames = [];		// collect all checked keys here

	with (dialog)
	{
		progress = -1;
		progressBar.repaint();

		for (var i = 0; i < keyBox.numberOfChildren; i++)
		{
			var node = keyBox.child(i);
			if (node.checked) keyNames.push(node.text(0));
		}
		//
		// write header
		//
		this.header.push("Filename");
		//
		// add checked keys
		//
		for (var i in keyNames) this.header.push(keyNames[i]);
		//
		// add checked functions
		//
		for (var i = 0; i < sdBox.numberOfChildren; i++)
		{
			var node = sdBox.child(i);
			if (node.checked) this.header.push(node.text(0));
		}
      //
      // create help vectors
      //
      this.types = [];
      for (var i = 0; i < this.header.length; i++)
      {
         this.types.push(-1);
      }
      //
      // assemble the table items
      //
		for (i = 0; i < filesCollection.length; i++)
		{
         // types 0 = string, 1 = int, 2 = double, 3 = bool, 4 = others

         var index = 0;       // 1st column
	      var row   = [];      // clear row
			var fo    = filesCollection[i];
			//
			// protocol, show progress
			//
			Console.writeln(i + ": processing " + fo.name);
			Console.flush();
			//
			// begin the new row with the filename
			//
         row.push(fo.name);
         this.types[index] = 0;
         index +=1;
			//
			// append all selected keys
			//
			for (var j in keyNames)
			{
				var name = keyNames[j];
 				var v = fo.getValue(name);

            if (typeof(v) == "string") v = v.trim();

            row.push(v);

            if (typeof(v) == "number")
            {
               var isInt = Math.frac(v) == 0;
               if (this.types[index] == -1)        // first row
               {
                  if (isInt) this.types[index] = 1; else  this.types[index] = 2;
               }
               else if (!isInt) this.types[index] = 2;
            }
            else if (typeof(v) == "string")    this.types[index] == 0;
            else if (typeof(v) == "boolean")   this.types[index] == 3;
            else if (typeof(v) == "undefined") this.types[index] == 4;
            index +=1;
         }
			//
			// append all selected functions
			//
			for (var j = 0; j < sdBox.numberOfChildren; j++)
			{
				var node = sdBox.child(j);
				if (node.checked)
				{
					var functionName = node.text(1);
					var v = evaluateFunction(functionName, fo);
               row.push(v)

               if (typeof(v) == "number")
               {
                  var isInt = Math.frac(v) == 0;
                  if (this.types[index] == -1)        // first row
                  {
                     if (isInt) this.types[index] = 1; else  this.types[index] = 2;
                  }
                  else if (!isInt) this.types[index] = 2;
               }
               else if (typeof(v) == "string")    this.types[index] == 0;
               else if (typeof(v) == "boolean")   this.types[index] == 3;
               else if (typeof(v) == "undefined") this.types[index] == 4;
               index +=1;
				}
			}
			//
			// row ready
			//
			this.table.push(row);

			progress = i;
			progressBar.repaint();
      }
      //return this.rows;
   }
}

function fileObj(filename, dialog)
{
	this.name  = filename;
	this.shortName = File.extractNameAndExtension(filename);
	this.image = new Image();
	this.dialog = dialog;
	this.error = false;

	var suffix = File.extractExtension(filename).toLowerCase();

	var fmt = new FileFormat(suffix, true, false);

	if (fmt.isNull) return;
	//
	// precheck fits, read one block from header
	//
	var buffer = new ByteArray();
	try
	{
		this.error = true;
		var file = new File(filename);
		buffer = file.read( DataType_ByteArray, 2880 );
		file.close();
		this.error = (buffer.length < 2880);
		Console.writeln("buffer.length "+buffer.length);
	}
	catch (ex)
	{
		new MessageBox("Reading header failed: " + ex.message + "\r\n" + filename).execute();
	}

	if (this.error) return;

 	try
	{
		this.error = true;

		var ffi = new FileFormatInstance(fmt);

		ffi.open(filename);
		ffi.readImage(this.image);
		this.keys = ffi.keywords;		// FITSKeyword type
		ffi.close();

		this.error = false;
	}
	catch (ex)
	{
		new MessageBox("Reading image failed: " + ex.message + "\r\n" + filename).execute();
	}

	if (this.error) return;

	var tmpW = new ImageWindow(this.image.width, this.image.height, this.image.numberOfSelectedChannels );
	tmpW.mainView.beginProcess(UndoFlag_NoSwapFile);
	tmpW.mainView.image.apply(this.image);
	tmpW.mainView.endProcess();
   tmpW.zoomToOptimalFit();
   tmpW.currentView.setPropertyValue( "dispose", true );
	this.window = tmpW;

	// default statistic values

	this.statistics = new ImageStatistics(this.image);

	// adv. image analysis

	var fwh  = [];
	var resi = [];
	var ecce = [];
	var starcount = -1;

 	this.FWHM = function()
	{
		if (fwh.length == 0 && !this.error) advAlaysis(this.window);
		return fwh;
	}

	this.Eccentricities = function()
	{
		if (ecce.length == 0 && !this.error) advAlaysis(this.window);
		return ecce;
	}

	this.Residuals = function()
	{
		if (resi.length == 0 && !this.error) advAlaysis(this.window);
		return resi;
	}

	this.starCount = function()
	{
		if (starcount < 0 && !this.error) advAlaysis(this.window);
		return starcount;
	};

	function advAlaysis(window)
	{
		parameters.upperLimit = 0.95;
		var a = FWHMEccentricityResidualOfStarsImageWindow(window);

		fwh  = a[0];
		ecce = a[1];
		resi = a[2];
		starcount = Math.max(a[3].length, 0);
	}

	this.getValue = function(name)
	{
		for (var i in this.keys)
		{
			if (this.keys[i].name == name)
		{
         if (this.keys[i].isString)
         {
            var trimChar = '\'';
            var v = this.keys[i].value;
            if (v.length > 1)
               if (v.charAt(v.length - 1) == trimChar) v = v.substr(0, v.length - 1);
            if (v.length > 1)
               if (v.charAt(0) == trimChar) v = v.substring(1);
            return v;
         }
         else
            return this.keys[i].numericValue;
			}
		}
		return "?";
	}

	this.compare = function(filename)
	{
		if (filename == this.name) 		return true;
		if (filename == this.shortName)  return true;
		return false;
	}
}

function enablePage1Buttons(ctl)
{
	// enable/disable buttons
	with (ctl)
	{
		btnSelAll.enabled = filesBox.numberOfChildren > 0;
		btnInvertSel.enabled = filesBox.selectedNodes.length > 0;
		btnRemove.enabled = filesBox.selectedNodes.length > 0;
		btnClear.enabled = filesBox.numberOfChildren > 0;
		cbFullNames.enabled = filesBox.numberOfChildren > 0;
		btnEval.enabled = filesBox.numberOfChildren > 0;
	}
}


function FDVDialog()
{
   // establish dialog base

   this.__base__ = Dialog;
   this.__base__();

   var dlg = this;

	this.column = 0;
	this.filesCollection  = [];
	this.keyCollection = [];

	this.table = null;       // array with rows, columns
   this.tableHeader = [];
   this.tableSorts  = [];
   this.tableTypes  = [];  // 0 = string, 1 = int, 2 = double, 3 = bool, 4 = others

	this.numTrans = 1;
	this.progress = -1;
	this.recentKeys = [];
	this.recentFunc = [];
   this.precision  = -1;
   this.fullNames  = false;

   this.csvSeparator = '\t';
   this.stringLimiter = '';

	var recentKeysString  = Settings.read(SETKEY + "/Keys", DataType_String);
	var recentFuncString  = Settings.read(SETKEY + "/Func", DataType_String);
	var recentTransString = Settings.read(SETKEY + "/TransVal", DataType_String);
	this.precision        = Settings.read(SETKEY + "/Precision", DataType_Int16);

   if (this.precision == null) this.precision = 8;

	if (recentKeysString != null)	this.recentKeys = recentKeysString.split('\t');
	if (recentFuncString != null)	this.recentFunc = recentFuncString.split('\t');
	if (recentTransString != null) this.numTrans  = parseInt(recentTransString);


	// make a form

	this.page1  = new GroupBox(this);
	with (this.page1)
	{
		title = "Add Files";

		this.filesBox = new TreeBox(this);
		with (this.filesBox)
		{
			//
			// create a treebox with 2 columns
			// column 0 contains filenames
			// column 1 contains the full path
			//
			alternateRowColor = true;
			numberOfColumns = 2;
			rootDecoration = false;
			setMinWidth(400);
			multipleSelection = true;
			setHeaderText(0, "FileNames");
			setHeaderText(1, "FileNames");
			hideColumn(1);

			onNodeClicked = function( item, columnIndex )
			{
			 	enablePage1Buttons(this.dialog);
			}

			onNodeDoubleClicked = function( item, columnIndex )
			{
				if (columnIndex == 0)
				{
					var filename = item.text(0);
					for (var i in this.dialog.filesCollection)
					{
						if (this.dialog.filesCollection[i].compare(filename))
						{
							viewFileObject(this.dialog.filesCollection[i]);
							break;
						}
					}
				}
			}

			onNodeExpanded = function( item, columnIndex )
			{
				enablePage1Buttons(this.dialog);
			}
		}
   }

   this.btnAdd = new PushButton(this);
   with (this.btnAdd)
   {
      text = "Add Files";

      onClick  = function()
      {
         var ofd = new OpenFileDialog();
         ofd.multipleSelections = true;
         ofd.filters = [ [ "FITS Files", ".fit", ".fits", ".fts" ],
                         [ "XISF Files", "*.xisf"]];

         if (ofd.execute())
         {
            for (var i = 0; i < ofd.fileNames.length; i++)
            {
               var filename =  ofd.fileNames[i];
               //
               // find filename in treebox column 1
               //
               var index = -1;
               for (var j = 0; j < this.dialog.filesBox.numberOfChildren; j++)
               {
                  if (this.dialog.filesBox.child(j).text(1) == filename)
                  {
                     index = j;
                     break;
                  }
               }
               if (index < 0)
               {
                  this.dialog.btnWriteCSV.enabled = false;
                  // add, if not found
                  addToCollection(filename, this.dialog);
               }
            }
         }
         this.dialog.progress = -1;
      }
   }

   this.btnSelAll = new PushButton(this);
   with (this.btnSelAll)
   {
      enabled = false;
      text = "Select All";

      onClick  = function()
      {
         this.dialog.filesBox.selectAllNodes();
         enablePage1Buttons(this.dialog);
      }
   }

   this.btnInvertSel = new PushButton(this);
   with (this.btnInvertSel)
   {
      enabled = false;
      text = "Invert Selection";

      onClick  = function()
      {
         for (var i = 0; i < this.dialog.filesBox.numberOfChildren; i++)
         {
            var n = this.dialog.filesBox.child(i);
            n.selected = !n.selected;
         }
         enablePage1Buttons(this.dialog);
      }
   }

   this.btnRemove = new PushButton(this);
   with (this.btnRemove)
   {
      enabled = false;
      text = "Remove Selected";

      onClick  = function()
      {
         var nodes  = this.dialog.filesBox.selectedNodes;
         for (var i = 0; i < nodes.length; i++)
         {
            var name = nodes[i].text(1);
            var c = this.dialog.filesBox.childIndex(nodes[i]);
            this.dialog.filesBox.remove(c);
            var index = -1;
            for (var j = 0; j < this.dialog.filesCollection.length; j++)
            {
               if (this.dialog.filesCollection[j].name == name)
               {
                  // force deletion
                  this.dialog.filesCollection[j].error = true;
                  index = j;
                  break;
               }
            }
         }
         var tmp = [];
         for (var j = 0; j < this.dialog.filesCollection.length; j++)
         {
            if (this.dialog.filesCollection[j].error) continue;
            tmp.push(this.dialog.filesCollection[j]);
         }
         this.dialog.filesCollection = [];
         for (var j = 0; j < tmp.length; j++)
            this.dialog.filesCollection.push(tmp[j]);
         tmp = [];
         enablePage1Buttons(this.dialog);
         this.dialog.txtCount.text = this.dialog.filesCollection.length.toString();
         this.dialog.progress = -1;
      }
   }

   this.btnClear = new PushButton(this);
   with (this.btnClear)
   {
      enabled = false;
      text = "Clear";

      onClick  = function()
      {
         this.dialog.filesBox.clear();
         for (var i in this.dialog.filesCollection)
         {
            if (!this.dialog.filesCollection[i].error)
               this.dialog.filesCollection[i].window.forceClose();
         }
         this.dialog.filesCollection = [];
         enablePage1Buttons(this.dialog);
         this.dialog.txtCount.text = this.dialog.filesCollection.length.toString();
         this.dialog.progress = -1;
      }
   }

   this.cbFullNames = new CheckBox(this);
   with (this.cbFullNames)
   {
      enabled = false;
      text = "Full names";

      onCheck = function(checked)
      {
         this.dialog.filesBox.hideColumn(this.dialog.column);
         this.dialog.column = 1 - this.dialog.column;
         this.dialog.filesBox.showColumn(this.dialog.column);
         this.dialog.fullNames = checked;
      }
   }

   this.lblCount = new Label(this);
   this.lblCount.text = "Count";

   this.txtCount = new Edit(this);
   with (this.txtCount)
   {
      backgroundColor  = 0xFFFFFFFF;
      readOnly = true;
      rightAlignment = true;
      setMaxSize(width / 2, height);
   }

   var countSizer = new HorizontalSizer();
   with (countSizer)
   {
      margin = 4;
      spacing = 6;
      add (this.lblCount);
      add (this.txtCount)
   }

   var leftPanel1 = new VerticalSizer();
   with (leftPanel1)
   {
      margin = 4;
      spacing = 6;
      add(this.filesBox);
   }

   var rightPanel1 = new VerticalSizer();
   with (rightPanel1)
   {
      margin = 4;
      spacing = 6;
      add(this.btnAdd);
      add(this.btnSelAll);
      add(this.btnInvertSel);
      add(this.btnRemove);
      add(this.btnClear);
      add(this.cbFullNames);
      add(countSizer);
      addStretch();
   }

   this.page1.sizer = new HorizontalSizer();
   with (this.page1.sizer)
   {
      margin = 4;
      spacing = 6;
      add(leftPanel1);
      add(rightPanel1);
   }


// *****************************************************************************

	this.page2  = new GroupBox(this);
	with (this.page2)
	{
		title = "Fits Key Selection";

		this.keyBox = new TreeBox(this);
		with (this.keyBox)
		{
			//
			// create a treebox with 1 column containing FITS keys
			//
			alternateRowColor = true;
			numberOfColumns = 1;
			rootDecoration = false;
			multipleSelection = true;
			setHeaderText(0, "Keys");
			setMaxWidth(150);


			onNodeUpdated = function( item, columnIndex )
			{
				this.dialog.btnWriteCSV.enabled = false;

				var i = this.dialog.recentKeys.indexOf(item.text(0));
				if (item.checked)
				{
					if (i < 0) this.dialog.recentKeys.push(item.text(0));
				}
				else
				{
					if (i > -1)
					{
						this.dialog.recentKeys[i] = "";
						this.dialog.recentKeys.sort();
						this.dialog.recentKeys.pop();
					}
				}
			}
		}

		this.btnCheckAll = new PushButton(this);
		with (this.btnCheckAll)
		{
			text = "Check All";

			onClick  = function()
			{
				for (var i = 0; i < this.dialog.keyBox.numberOfChildren; i++)
				{
					var node = this.dialog.keyBox.child(i);
					node.checked = true;
				}
				this.dialog.recentKeys = [];
			}
		}

		this.btnUnCheckAll = new PushButton(this);
		with (this.btnUnCheckAll)
		{
			text = "Uncheck All";

			onClick  = function()
			{
				for (var i = 0; i < this.dialog.keyBox.numberOfChildren; i++)
				{
					var node = this.dialog.keyBox.child(i);
					node.checked = false;
				}
				this.dialog.recentKeys = [];
			}
		}

		var leftPanel2 = new HorizontalSizer();
		with (leftPanel2)
		{
      	margin = 4;
      	spacing = 6;
			add(this.keyBox);
		}

		var rightPanel2 = new VerticalSizer();
		with (rightPanel2)
		{
			margin = 4;
      	spacing = 6;
			add(this.btnCheckAll);
			add(this.btnUnCheckAll);
			addStretch();
		}

      sizer = new HorizontalSizer();
      sizer.margin = 4;
      sizer.spacing = 6;
      sizer.add(leftPanel2);
      sizer.add(rightPanel2);
      sizer.addStretch();
	}

// *****************************************************************************

	this.page3  = new GroupBox(this);
	with (this.page3)
	{
		title = "Statistical Data Selection";

		this.sdBox = new TreeBox(this);
		with (this.sdBox)
		{
			//
			// create a treebox with 2 columns
			// column 0 contains the data description
			// column 1 contains a local function name for this element
			//
			alternateRowColor = true;
			numberOfColumns = 2;
			rootDecoration = false;
			multipleSelection = false;
			setHeaderText(0, "Statistical Data Elements");
 			setMaxWidth(450);
			showColumn(0, true);
			showColumn(1, false);
         setColumnWidth(0, 300);
			fillFunctions(this.sdBox);

			try
			{
				for (var i = 0; i < this.dialog.sdBox.numberOfChildren; i++)
				{
					var node = this.dialog.sdBox.child(i);
					for (var j = 0; j < this.dialog.recentFunc.length; j++)
					{
						var func = this.dialog.recentFunc[j];
						var funcNameFlag = func.split('|');
						if (funcNameFlag.length == 2)
						{
							if (funcNameFlag[0] == node.text(1))
							{
								node.checked = (funcNameFlag[1] == "true");
								break;
							}
						}
					}
				}
			}
			catch (ex)
			{
				new MessageBox("Init failed: " + ex.message).execute();
			}

			onNodeClicked = function( item, columnIndex )
			{
				if (columnIndex == 2)
					this.dialog.cmbNumBase.showList();
			}
		}

		this.btnCheckAll3 = new PushButton(this);
		with (this.btnCheckAll3)
		{
			text = "Check All";

			onClick  = function()
			{
				for (var i = 0; i < this.dialog.sdBox.numberOfChildren; i++)
				{
					var node = this.dialog.sdBox.child(i);
					node.checked = true;
				}
			}
		}

		this.btnUnCheckAll3 = new PushButton(this);
		with (this.btnUnCheckAll3)
		{
			text = "Uncheck All";

			onClick  = function()
			{
				for (var i = 0; i < this.dialog.sdBox.numberOfChildren; i++)
				{
					var node = this.dialog.sdBox.child(i);
					node.checked = false;
				}
			}
		}

		this.cmbNumBase = new ComboBox();
		with (this.cmbNumBase)
		{
			addItem("Normalized Real [0,1]");
			addItem("8 bit [0,255]");
			addItem("10 bit [0,1023]");
			addItem("12 bit [0,4095]");
			addItem("14 bit [0,16383]");
			addItem("16 bit [0,65535]");
 			//
			// old value to item
			//
			switch (this.dialog.numTrans)
			{
				case 1     : currentItem = 0; break;
				case 255   : currentItem = 1; break;
				case 1023  : currentItem = 2; break;
				case 4095  : currentItem = 3; break;
				case 16383 : currentItem = 4; break;
				case 65535 : currentItem = 5; break;
			}

			onItemSelected = function( itemIndex )
			{
				if (this.dialog.sdBox.currentNode != null)
				{
					switch (itemIndex)
					{
						case 0 : this.dialog.numTrans = 1; break;
						case 1 : this.dialog.numTrans = 255; break;
						case 2 : this.dialog.numTrans = 1023; break;
						case 3 : this.dialog.numTrans = 4095; break;
						case 4 : this.dialog.numTrans = 16383; break;
						case 5 : this.dialog.numTrans = 65535; break;
					}
					// copy into file objects
					for (var i in this.dialog.filesCollection)
					 	this.dialog.filesCollection[i].numTrans = this.dialog.numTrans;
				}
			}
		}

		var leftPanel3 = new HorizontalSizer();
		with (leftPanel3)
		{
      	margin = 4;
      	spacing = 6;
			add(this.sdBox);
		}

		var rightPanel3 = new VerticalSizer();
		with (rightPanel3)
		{
			margin = 4;
      	spacing = 6;
			add(this.btnCheckAll3);
			add(this.btnUnCheckAll3);
			add(this.cmbNumBase);
			addStretch();
		}

      sizer = new HorizontalSizer();
      sizer.margin = 4;
      sizer.spacing = 6;
      sizer.add(leftPanel3);
      sizer.add(rightPanel3);
	}

// *****************************************************************************

	this.page4  = new GroupBox(this);
	with (this.page4)
	{
		title = "Report";

		this.btnEval = new PushButton(this);
		with (this.btnEval)
		{
			enabled = false;
			text ="Generate report";

			onClick = function()
			{
				enabled = false;
				this.dialog.btnPreview.enabled = false;
				this.dialog.btnWriteCSV.enabled = false;

            var tg = new tableGenerator(dialog);

            this.dialog.table  = tg.table;
            this.dialog.header = tg.header;
            this.dialog.types  = tg.types;
				this.dialog.btnPreview.enabled = true;
				this.dialog.btnWriteCSV.enabled = true;
				enabled = true;
			}
		}

		this.progressBar = new Label(this);
		with (this.progressBar)
		{
			backgroundColor = 0xFFF0F0F0;
			scaledMinHeight = 24;
			textAlignment = TextAlign_HorzCenter|TextAlign_VertCenter;

			onPaint = function()
			{
				var i = this.dialog.progress;
 				var progressPercent = (i + 1) / this.dialog.filesCollection.length * 100;
 				var g      = new Graphics();
				g.begin(this.dialog.progressBar);
				g.clipRect = new Rect(0, 0, width, height);
				if (progressPercent == 0)
				{
					var clear = new Brush( parent.backgroundColor );
					g.clipRect = new Rect(0, 0, width, height);
					g.fillRect( g.clipRect, clear );
					this.dialog.progressBar.text =  (i + 1) + " (" + this.dialog.filesCollection.length + ")";
				}
				else
				{
					var clear = new Brush( 0xFFFFFFFF );
					var r = new Rect(0, 0, parent.width - position.x * 2, height);
					g.clipRect = r;
					g.fillRect( g.clipRect, clear );
					var w = progressPercent * r.width * 0.01;
					r = new Rect(0, 0, w, height);
					g.fillRect(r, new Brush(0xFF00FF00));
					this.dialog.progressBar.text =  (i + 1) + " (" + this.dialog.filesCollection.length + ")";
				}
				g.end();
			}
		}

      this.btnPreview = new PushButton(this);
      with (this.btnPreview)
      {
         enabled = false;
         text = "Preview";

         onClick = function()
         {
            var pvd = new PreviewDialog(this.dialog);
            pvd.execute();
         }
      }


      this.lblSeparator = new Label(this);
      this.lblSeparator.text = 'Data Separator:';

      this.cmbSeparators = new ComboBox(this);
      with (this.cmbSeparators)
      {
         addItem('TAB');
         addItem(',');
         addItem(';');
         addItem('|');
         addItem('space');

         toolTip = "<p>Set csv data separator.</p>";

         onItemSelected = function ( itemIndex )
         {
            switch (itemIndex)
            {
               case 0:
               {
                  dlg.csvSeparator = '\t';
                  break;
               }
               case 1:
               {
                  dlg.csvSeparator = ',';
                  break;
               }
               case 2:
               {
                  dlg.csvSeparator = ';';
                  break;
               }
               case 3:
               {
                  dlg.csvSeparator = '|';
                  break;
               }
               case 4:
               {
                  dlg.csvSeparator = ' ';
                  break;
               }
            }
         }
      }

      this.lblLimiter = new Label(this);
      this.lblLimiter.text = 'String Limiter:';

      this.cmbLimiters = new ComboBox(this);
      with (this.cmbLimiters)
      {
         addItem('none');
         addItem('\'');
         addItem('\"');

         toolTip = "<p>Set csv string limiter.</p>";

         onItemSelected = function ( itemIndex )
         {
            switch (itemIndex)
            {
               case 0:
               {
                  dlg.stringLimiter= '';
                  break;
               }
               case 1:
               {
                  dlg.stringLimiter = '\'';
                  break;
               }
               case 2:
               {
                  dlg.stringLimiter = '\"';
                  break;
               }
            }
         }
      }


		this.btnWriteCSV = new PushButton(this);
		with (this.btnWriteCSV)
		{
			enabled = false;
			text ="Write csv-file";

			onClick = function()
			{
            var csvSeparator = dlg.csvSeparator;
            var stringLimiter = dlg.stringLimiter;

				var d = new SaveFileDialog();
				d.overwritePrompt = true;
				d.caption = "Save FITS report";
				d.filters = [['CSV file', '*.csv']];

				if (d.execute())
				{
					try
					{
						var file = new File();
						file.create(d.fileName);

                  var items = [];
                  for (var i in this.dialog.header)
                  {
                     var s = this.dialog.header[i];
                     if (isNaN(s) & csvSeparator == ' ')
                        s = stringLimiter + s + stringLimiter;
                     items.push(s);
                  }

                  file.outTextLn(items.join(csvSeparator));

						for (var i = 0; i < this.dialog.table.length; i++)
						{
                     var row = this.dialog.table[i];
                     items = [];
                     for (var j in row)
                     {
                        var s = "";
                        var v = row[j];

                        if (this.dialog.types[j] == 0)
                        {
                           if (j == 0 && !this.dialog.fullNames)
                              v = File.extractNameAndExtension(v);
                           s = stringLimiter + v + stringLimiter;
                        }
                        else s = v.toString();

                        items.push(s);
                     }
							file.outTextLn(items.join(csvSeparator));
						}
						file.close();
						new MessageBox("Report saved as " + d.fileName).execute();
					}
					catch (ex)
					{
						new MessageBox("Write failed, report not saved: " + ex.message).execute();
					}
				}
			}
		}


      this.frameSeparator = new Frame(this);
      with (this.frameSeparator)
      {
         height = this.btnWriteCSV.height;
         width = this.btnWriteCSV.width;
   
         sizer = new HorizontalSizer();
         with (sizer)
         {
            margin = 0;
            spacing = 12;
            add(this.lblSeparator);
            addStretch();
            add(this.cmbSeparators);
         }
      }


      this.frameLimiter = new Frame(this);
      with (this.frameLimiter)
      {
         height = this.btnWriteCSV.height;
         width = this.btnWriteCSV.width;
   
         sizer = new HorizontalSizer();
         with (sizer)
         {
            margin = 0;
            spacing = 12;
            add(this.lblLimiter);
            addStretch();
            add(this.cmbLimiters);
         }
      }

      
      this.btnQuit = new PushButton(this);
		with (this.btnQuit)
		{
			text = "Quit";

			onClick = function()
			{
				//
				// save key and function selection
				//
				var keyNames = [];
				for (var i = 0; i < this.dialog.keyBox.numberOfChildren; i++)
				{
					var node =this.dialog. keyBox.child(i);
					if (node.checked) keyNames.push(node.text(0));
				}
				var checkedKeysString = keyNames.join('\t');
				var functionNames = [];
				for (var i = 0; i < this.dialog.sdBox.numberOfChildren; i++)
				{
					var node = this.dialog.sdBox.child(i);
					functionNames.push( node.text(1) + '|' + node.checked  );
				}
				var functionNamesString = functionNames.join('\t');

				Settings.write(SETKEY + "/Keys", DataType_String, checkedKeysString);
				Settings.write(SETKEY + "/Func", DataType_String, functionNamesString);
				Settings.write(SETKEY + "/TransVal", DataType_String, this.dialog.numTrans.toString());
				Settings.write(SETKEY + "/Precision", DataType_Int16, this.dialog.precision);


            var listFilenames = [];
            for (var i in this.dialog.filesCollection)
            {
              listFilenames.push(this.dialog.filesCollection[i].name);
            }
            var strFilenames = listFilenames.join('\r');
            Settings.write(SETKEY + "/Files", DataType_String, strFilenames);

				this.dialog.done(0);
			}
		}

		// my ©

		this.lblCopyright = new Label(this)
   	this.lblCopyright.text = "© 2018, Hartmut V. Bornemann";

 		var centerCol = new VerticalSizer();

		with (centerCol)
		{
			margin = 4;
      	spacing = 6;
			add(this.btnEval);
			add(this.progressBar);
         add(this.btnPreview);
         addSpacing(8);
         add(this.frameSeparator);
         add(this.frameLimiter);
      	add(this.btnWriteCSV);
         addSpacing(8);
			addStretch();
			add(this.btnQuit);
			addStretch();
			add(this.lblCopyright);
		}

		sizer = new HorizontalSizer(this);
		with (sizer)
		{
			margin = 4;
      	spacing = 6;
			addStretch();
			add(centerCol);
			addStretch();
		}
	}

// *****************************************************************************


// *****************************************************************************


	this.tabBox = new TabBox(this);
	with (this.tabBox)
	{
		addPage(parent.page1, "Fits Files");
		addPage(parent.page2, "Fits Keys");
		addPage(parent.page3, "Image Statistics");
		addPage(parent.page4, "Create Report");
		setVariableSize();
	}
	this.sizer= new HorizontalSizer();
   this.sizer.margin = 4;
   this.sizer.add(this.tabBox);
   this.windowTitle = TITLE + " " + VERSION;
   this.adjustToContents();
	//this.userResizable = false;

   var recentFilesString = Settings.read(SETKEY + "/Files", DataType_String);
	if (recentFilesString != null)
   {
      //Console.writeln(recentFilesString, recentFilesString.length);
      var listFiles = recentFilesString.split('\r');
      for (var i in listFiles)
      {
         //if (listFiles[i].length > 0) addToCollection(listFiles[i], this.dialog);
      }
   }
   enablePage1Buttons(this);


   function addToCollection(filename, dialog)
   {
      var fo = new fileObj(filename, dialog);
      if (!fo.error)
      {
         var node = new TreeBoxNode(dialog.filesBox);
         node.setText(0, fo.shortName);
         node.setText(1, fo.name);
         Console.writeln("Read ok: " + filename);
         dialog.filesCollection.push(fo);
         dialog.txtCount.text = dialog.filesCollection.length.toString();
         for (var j = 0; j < fo.keys.length; j++)
         {
            var keyWord = fo.keys[j];	// FITSKeyword type
            if (keyWord.name == "COMMENT") continue;
            if (keyWord.name == "CONTINUE") continue;
            if (keyWord.name == "HISTORY") continue;
            var k = dialog.keyCollection.indexOf(keyWord.name);
            if (k < 0)
            {
               dialog.keyCollection.push(keyWord.name);
               node = new TreeBoxNode(dialog.keyBox);
               node.checkable = true;
               node.checked = false;
               node.setText(0, keyWord.name);
            }
         }
         //
         // check recent keys
         //
         for (var i in dialog.recentKeys)
         {
            for (var j = 0; j < dialog.keyBox.numberOfChildren; j++)
            {
               var node = dialog.keyBox.child(j);
               var key = node.text(0);
               if (key == dialog.recentKeys[i])
                  node.checked = true;
            }
         }
      }
      else
         Console.writeln("Read error: " + filename);

      enablePage1Buttons(dialog);
   }
}

function addFunction(tree, description, functionName)
{
	var node = new TreeBoxNode(tree);
	node.checkable = true;
	node.checked = false;
	node.setText(0, description);
	node.setText(1, functionName);
	return true;
}

function SNREvaluation ( img )
{
   var E = new NoiseEvaluation( img );
   var snr = img.meanOfSquares(img.selectedRect)/ E.sigma / E.sigma;
   var db = 10 * Math.log10(snr);
   return db;
}

function NoiseEvaluation( img )
{
   var a, n = 4, m = 0.01*img.selectedRect.area;
   for ( ;; )
   {
      a = img.noiseMRS( n );
      if ( a[1] >= m )
         break;
      if ( --n == 1 )
      {
         console.writeln( "<end><cbr>** Warning: No convergence in MRS noise evaluation routine - using k-sigma noise estimate." );
         a = img.noiseKSigma();
         break;
      }
   }
   this.sigma = a[0]; // estimated stddev of Gaussian noise
   this.count = a[1]; // number of pixels in the noise pixels set
   this.layers = n;   // number of layers used for noise evaluation
}

// This section is part of SubframeSelector Script version 1.11
//
// Copyright (C) 2012-2016 Mike Schuster. All Rights Reserved.
// Copyright (C) 2003-2016 Pleiades Astrophoto S.L. All Rights Reserved.

function starDescription(b, a, x, y, sx, sy, theta, residual) {
   this.b = b;
   this.a = a;
   this.x = x;
   this.y = y;
   this.sx = sx;
   this.sy = sy;
   this.theta = theta;
   this.residual = residual;
}

function starDescriptionCompare(a, b) {
   var ax = Math.round(a.x);
   var ay = Math.round(a.y);
   var bx = Math.round(b.x);
   var by = Math.round(b.y);
   return ax < bx ? -1 : ax > bx ? 1 : ay < by ? -1 : ay > by ? 1 : 0;
}

function uniqueViewIdNoLeadingZero(baseId) {
   var id = baseId;
   for (var i = 1; !View.viewById(id).isNull; ++i) {
      id = baseId + format("%d", i);
   }
   return id;
}

function uniqueArray(values, compareFunction) {
   if (values.length < 2) {
      return values;
   }
   values.sort(compareFunction);

   var j = 0;
   for (var i = 1; i != values.length; ++i) {
      if (compareFunction(values[j], values[i]) == -1) {
         ++j;
         values[j] = values[i];
      }
   }
   return values.slice(0, j + 1);
}

function FWHMEccentricityResidualOfStarsImageWindow(imageWindow) {
   var starsViewId = uniqueViewIdNoLeadingZero(imageWindow.mainView.id + "_stars");

   var barycenters = new Array;
   var threshold = 1.0;
   var maximumDetectedStars = 100000;
   if (parameters.upperLimit == 0.0) {
   }
   else if (USE_STAR_DETECTOR) {
      console.writeln();
      console.writeln("<b>StarDetector ", #__PJSR_STAR_DETECTOR_VERSION,
					"</b>: Processing view: " + imageWindow.mainView.id   );
      console.flush();
      var startTime = new Date();

      var starDetector = new StarDetector();
      starDetector.structureLayers = parameters.starDetectionLayers;
      starDetector.noiseLayers = Math.min(
         parameters.noiseReductionLayers, parameters.starDetectionLayers - 1
      );
      starDetector.hotPixelFilterRadius = parameters.hotPixelFilterRadius;
      starDetector.applyHotPixelFilterToDetectionImage = parameters.applyHotPixelFilterToDetectionImage;
      starDetector.sensitivity = parameters.starDetectionSensitivity;
      starDetector.peakResponse = parameters.starPeakResponse;
      starDetector.maxDistortion = parameters.maximumStarDistortion;
      starDetector.upperLimit = parameters.upperLimit;

      var stars = starDetector.stars(imageWindow.mainView.image);

      var endTime = new Date();
      console.writeln(stars.length, " star(s) found");
      console.writeln(format("%.03f s", 0.001 * (endTime.getTime() - startTime.getTime())));
      console.flush();

      for (;;) {
         for (var i = 0; i != stars.length && barycenters.length != maximumDetectedStars; ++i) {
            if (threshold == 1.0 || Math.random() <= threshold) {
               barycenters.push({
                  position: stars[i].pos,
                  radius: Math.max(3, Math.ceil(Math.sqrt(stars[i].size)))
               });
            }
         }
         if (barycenters.length != maximumDetectedStars) {
            break;
         }
         barycenters = new Array;
         threshold = 0.1 * threshold;
      }
   }
   else {
      var starAlignment = new StarAlignment;
      with (starAlignment) {
         referenceImage = imageWindow.mainView.id;
         referenceIsFile = false;

         mode = StarAlignment.prototype.DrawStars;
         onError = StarAlignment.prototype.Continue;

         structureLayers = parameters.starDetectionLayers;
         noiseLayers = Math.min(
            parameters.noiseReductionLayers, parameters.starDetectionLayers - 1
         );
         hotPixelFilterRadius = parameters.hotPixelFilterRadius;
         sensitivity = parameters.starDetectionSensitivity;
         peakResponse = parameters.starPeakResponse;
         maxStarDistortion = parameters.maximumStarDistortion;
      }
      //console.abortEnabled = false;
      starAlignment.executeOn(imageWindow.mainView);
      console.abortEnabled = true;

      var starsView = View.viewById(starsViewId);
      if (starsView.isNull) {
         console.writeln("");
         console.writeln("*** Error: StarAlignment star map not found: ", starsViewId);
         return [0.0, 0.0, 0];
      }
      var starsImageWindow = starsView.window;
      starsImageWindow.hide();

      var barycentersImageWindow = new ImageWindow(
         starsImageWindow.mainView.image.width,
         starsImageWindow.mainView.image.height,
         starsImageWindow.mainView.image.numberOfChannels,
         32,
         true,
         starsImageWindow.mainView.image.colorSpace != ColorSpace_Gray,
         uniqueViewId(imageWindow.mainView.id + "_barycenters")
      );

      starsImageWindow.createPreview(starsImageWindow.mainView.image.bounds, "_");
      with (barycentersImageWindow.mainView) {
         beginProcess(UndoFlag_NoSwapFile);
         with (image) {
            var barycenterKernel = [
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0
            ];
            selectedPoint = new Point(0, 0);
            apply(starsImageWindow.previewById("_").image);
            binarize(0.5);
            convolve(new Matrix(barycenterKernel, 11, 11));
            binarize(1.0);
            resetSelections();
         }
         endProcess();
      }
      starsImageWindow.deletePreview(starsImageWindow.previewById("_"));
      starsImageWindow.forceClose();

      var radius = Math.round(0.75 * (new DynamicPSF).searchRadius);
      var barycentersImage = barycentersImageWindow.mainView.image;
      for (;;) {
         for (
            var i = 0;
            i != barycentersImage.width && barycenters.length != maximumDetectedStars;
            ++i
         ) {
            for (
               var j = 0;
               j != barycentersImage.height && barycenters.length != maximumDetectedStars;
               ++j
            ) {
               if (
                  barycentersImage.sample(i, j) == 1.0 &&
                  barycentersImage.sample(i - 1, j - 1) +
                  barycentersImage.sample(i - 1, j) +
                  barycentersImage.sample(i - 1, j + 1) +
                  barycentersImage.sample(i, j - 1) +
                  barycentersImage.sample(i, j + 1) +
                  barycentersImage.sample(i + 1, j - 1) +
                  barycentersImage.sample(i + 1, j) +
                  barycentersImage.sample(i + 1, j + 1) == 0.0 &&
                  (threshold == 1.0 || Math.random() <= threshold)
               ) {
                  barycenters.push({
                     position: new Point(i, j),
                     radius: radius
                  });
               }
            }
         }
         if (barycenters.length != maximumDetectedStars) {
            break;
         }
         barycenters = new Array;
         threshold = 0.1 * threshold;
      }
      barycentersImageWindow.forceClose();
   }

   var maximumFittedStars = 20000;
   if (barycenters.length > maximumFittedStars) {
      threshold = 0.1 * threshold;
      for (var i = barycenters.length - 1; i != 0; --i) {
         var j = Math.round(Math.random() * i);
         var x = barycenters[i];
         barycenters[i] = barycenters[j];
         barycenters[j] = x;
      }
      barycenters = barycenters.slice(0, maximumFittedStars);
   }

   //if (threshold < 1.0) {
   //   console.writeln("");
   //   console.writeln("** Warning: excessive number of detected stars");
   //}

   var dynamicPSF = new DynamicPSF;
   with (dynamicPSF) {
      autoPSF = false;
      circularPSF = parameters.circularModel;
      gaussianPSF = parameters.modelFunction == 0;
      moffatPSF = false;
      moffat10PSF = parameters.modelFunction == 1;
      moffat8PSF = parameters.modelFunction == 2;
      moffat6PSF = parameters.modelFunction == 3;
      moffat4PSF = parameters.modelFunction == 4;
      moffat25PSF = parameters.modelFunction == 5;
      moffat15PSF = parameters.modelFunction == 6;
      lorentzianPSF = parameters.modelFunction == 7;
      regenerate = true;
   }

   var views = new Array;
   views.push(new Array(imageWindow.mainView.id));
   dynamicPSF.views = views;

   var radius = Math.round(1.5 * dynamicPSF.searchRadius);
   var stars = new Array;
   for (var i = 0; i != barycenters.length; ++i) {
      stars.push(new Array(
         0, 0, DynamicPSF.prototype.Star_DetectedOk,
         barycenters[i].position.x - barycenters[i].radius,
         barycenters[i].position.y - barycenters[i].radius,
         barycenters[i].position.x + barycenters[i].radius,
         barycenters[i].position.y + barycenters[i].radius,
         barycenters[i].position.x,
         barycenters[i].position.y
      ));
   }
   dynamicPSF.stars = stars;
   var fitted = new Array(stars.length);
   for (var i = 0; i != fitted.length; ++i) {
      fitted[i] = false;
   }

   //console.abortEnabled = false;
   // Workaround for 1195 Win 7 DynamicPSF failure on images with clipped backgrounds
   if (imageWindow.mainView.image.median() != 0) {
      dynamicPSF.executeGlobal();
   }
   else {
      console.writeln("0 PSF fittings");
   }
   console.abortEnabled = true;

   #define DYNAMICPSF_PSF_StarIndex 0
   #define DYNAMICPSF_PSF_Status 3
   #define DYNAMICPSF_PSF_b 4
   #define DYNAMICPSF_PSF_a 5
   #define DYNAMICPSF_PSF_cx 6
   #define DYNAMICPSF_PSF_cy 7
   #define DYNAMICPSF_PSF_sx 8
   #define DYNAMICPSF_PSF_sy 9
   #define DYNAMICPSF_PSF_theta 10
   #define DYNAMICPSF_PSF_mad 12

   var starDescriptions = new Array;
   var psfTable = dynamicPSF.psf;
   var starsTable = dynamicPSF.stars;
   for (var i = 0; i != psfTable.length; ++i) {
      var psfRow = psfTable[i];
      if (
         psfRow[DYNAMICPSF_PSF_Status] == DynamicPSF.prototype.PSF_FittedOk &&
         psfRow[DYNAMICPSF_PSF_mad] < parameters.maxPSFMAD &&
         !fitted[psfRow[DYNAMICPSF_PSF_StarIndex]]
      ) {
         var starsRow = starsTable[psfRow[DYNAMICPSF_PSF_StarIndex]];
         starDescriptions.push(new starDescription(
            psfRow[DYNAMICPSF_PSF_b],
            psfRow[DYNAMICPSF_PSF_a],
            psfRow[DYNAMICPSF_PSF_cx],
            psfRow[DYNAMICPSF_PSF_cy],
            psfRow[DYNAMICPSF_PSF_sx],
            psfRow[DYNAMICPSF_PSF_sy],
            psfRow[DYNAMICPSF_PSF_theta],
            psfRow[DYNAMICPSF_PSF_mad]
         ));
         fitted[psfRow[DYNAMICPSF_PSF_StarIndex]] = true;
      }
   }
   starDescriptions = uniqueArray(starDescriptions, starDescriptionCompare);

   var FWHMs = [];
   var eccentricities = [];
   var residuals = [];
   for (var i = 0; i != starDescriptions.length; ++i) {
      var description = starDescriptions[i];
      FWHMs.push(Math.sqrt(description.sx * description.sy));
      eccentricities.push(
         Math.sqrt(1.0 - Math.pow(description.sy / description.sx, 2.0))
      );
      residuals.push(description.residual);
   }

//Console.writeln("FWHM "+medianMeanDeviationOfArray(FWHMs));

   return [
      medianMeanDeviationOfArray(FWHMs),
      medianMeanDeviationOfArray(eccentricities),
      medianMeanDeviationOfArray(residuals),
      starDescriptions
   ];
}

function numberCompare(a, b) {
   return a < b ? -1 : a > b ? 1 : 0;
}

function medianMeanDeviationOfArray(values) {
   if (values.length == 0) {
      return [0.0, 0.0];
   }

   values.sort(numberCompare);

   var i = Math.floor(0.5 * values.length);
   var median = (2 * i == values.length) ?
      0.5 * (values[i - 1] + values[i]) :
      values[i];

   var deviations = new Array(values.length);
   for (var i = 0; i != values.length; ++i) {
      deviations[i] = Math.abs(median - values[i]);
   }

   var dispersion = 0.0;
   for (var i = 0; i != deviations.length; ++i) {
      dispersion += deviations[i];
   }
   dispersion = dispersion / deviations.length;

   return [median, dispersion];
}

function medianMedianDeviationOfArray(values) {
   if (values.length == 0) {
      return [0.0, 0.0];
   }

   values.sort(numberCompare);

   var i = Math.floor(0.5 * values.length);
   var median = (2 * i == values.length) ?
      0.5 * (values[i - 1] + values[i]) :
      values[i];

   var deviations = new Array(values.length);
   for (var i = 0; i != values.length; ++i) {
      deviations[i] = Math.abs(median - values[i]);
   }

   deviations.sort(numberCompare);

   var i = Math.floor(0.5 * deviations.length);
   var dispersion = (2 * i == deviations.length) ?
      0.5 * (deviations[i - 1] + deviations[i]) :
      deviations[i];

   return [median, dispersion];
}

function sigmaNormalize(value, median, dispersion) {
   return (value - median) / (dispersion != 0.0 ? dispersion : 1.0);
}

function sigmaDenormalize(value, median, dispersion) {
   return value * (dispersion != 0.0 ? dispersion : 1.0) + median;
}

function findMedianDispersionDescription(descriptions) {
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].weight);
   }
   var weight = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].FWHM);
   }
   var FWHM = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].FWHMMeanDeviation);
   }
   var FWHMMeanDeviation = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].eccentricity);
   }
   var eccentricity = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].eccentricityMeanDeviation);
   }
   var eccentricityMeanDeviation = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].SNRWeight);
   }
   var SNRWeight = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].median);
   }
   var median = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].noise);
   }
   var noise = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].dispersion);
   }
   var dispersion = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].starSupport);
   }
   var starSupport = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].starResidual);
   }
   var starResidual = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].starResidualMeanDeviation);
   }
   var starResidualMeanDeviation = medianMeanDeviationOfArray(values);
   var values = [];
   for (var i = 0; i != descriptions.length; ++i) {
      values.push(descriptions[i].noiseSupport);
   }
   var noiseSupport = medianMeanDeviationOfArray(values);

   return new evaluationDescription(
      null,
      true,
      false,
      false,
      0,
      weight,
      FWHM,
      FWHMMeanDeviation,
      eccentricity,
      eccentricityMeanDeviation,
      SNRWeight,
      median,
      noise,
      dispersion,
      starSupport,
      starResidual,
      starResidualMeanDeviation,
      noiseSupport,
      "",
      []
   );
}

function parametersPrototype() {
   this.scaleUnits = new Array(
      "Arcseconds (arcsec)",
      "Pixels (pixel)"
   );
   this.dataUnits = new Array(
      "Electrons (e-)",
      "Data Numbers (DN)"
   );

   this.cameraResolutions = new Array(
      "8-bit [0, 255]",
      "10-bit [0, 1023]",
      "12-bit [0, 4095]",
      "14-bit [0, 16383]",
      "16-bit [0, 65535]"
   );
   this.cameraResolutionValues = new Array(
      255,
      1023,
      4095,
      16383,
      65535
   );

   this.modelFunctions = new Array(
      "Gaussian",
      "Moffat 10",
      "Moffat 8",
      "Moffat 6",
      "Moffat 4",
      "Moffat 2.5",
      "Moffat 1.5",
      "Lorentzian"
   );
   this.modelScaleFactors = new Array(
      2.0 * Math.sqrt(2.0 * Math.log(2.0)), // Gaussian
      2.0 * Math.sqrt(Math.pow(2.0, 1.0 / 10.0) - 1.0), // Moffat 10
      2.0 * Math.sqrt(Math.pow(2.0, 1.0 / 8.0) - 1.0), // Moffat 8
      2.0 * Math.sqrt(Math.pow(2.0, 1.0 / 6.0) - 1.0), // Moffat 6
      2.0 * Math.sqrt(Math.pow(2.0, 1.0 / 4.0) - 1.0), // Moffat 4
      2.0 * Math.sqrt(Math.pow(2.0, 1.0 / 2.5) - 1.0), // Moffat 2.5
      2.0 * Math.sqrt(Math.pow(2.0, 1.0 / 1.5) - 1.0), // Moffat 1.5
      2.0 * Math.sqrt(Math.pow(2.0, 1.0 / 1.0) - 1.0) // Lorentzian
   );

   this.columnNames = new Array(
      "Index Name (0 approved/0)",
      "Weight",
      "FWHM (arcsec)",
      "Eccentricity",
      "SNRWeight",
      "Median (e-)",
      "MeanDeviation (e-)",
      "Noise (e-)",
      "StarSupport",
      "StarResidual",
      "NoiseSupport",
      "FWHMMeanDev (arcsec)",
      "EccentricityMeanDev",
      "StarResidualMeanDev",
      "Date (UTC)"
   );
   this.columnIndexNameApprovedTotal = function(approved, total) {
      return "Index Name (" +
         format("%d", approved) +
         " approved/" +
         format("%d", total) +
         ")";
   }
   this.columnName = function(index) {
      if (index == 2) {
         return parameters.scaleUnit == 0 ? "FWHM (arcsec)" : "FWHM (pixel)";
      }
      if (index == 5) {
         return parameters.dataUnit == 0 ? "Median (e-)" : "Median (DN)";
      }
      if (index == 6) {
         return parameters.dataUnit == 0 ? "MeanDeviation (e-)" : "MeanDeviation (DN)";
      }
      if (index == 7) {
         return parameters.dataUnit == 0 ? "Noise (e-)" : "Noise (DN)";
      }
      if (index == 11) {
         return parameters.scaleUnit == 0 ? "FWHMMeanDev (arcsec)" : "FWHMMeanDev (pixel)";
      }
      return this.columnNames[index];
   };
   this.columnToolTips = [
"<p>The <i>index</i> number and <i>file name</i> of the subframe in the target " +
"subframes list.</p>" +

"<p>See the tool tip messages of this row's columns for more information.</p>",

"<p>The <i>weight</i> of the subframe as determined by the " +
"subframe weighting expression.</p>",

"<p>The median star profile <i>full width at half maximum</i> (FWHM) " +
"estimate for the subframe in arcseconds or pixels.</p>",

"<p>The median star profile <i>eccentricity</i> estimate for " +
"the subframe.</p>",

"<p>The <i>signal to noise ratio weight</i> estimate for " +
"the subframe.</p>",

"<p>The <i>median</i> of the subframe in electrons or Data " +
"Numbers.</p>",

"<p>The <i>mean absolute deviation from the median</i> of " +
"the subframe in electrons or Data Numbers.</p>",

"<p>An estimate of the <i>standard deviation of Gaussian noise</i> " +
"for the subframe in electrons or Data Numbers.</p>",

"<p>The number of stars detected and fitted in the subframe and used to " +
"estimate <i>FWHM</i> and <i>Eccentricity</i>.</p>",

"<p>The median <i>residual</i> of the star fitting process " +
"for the subframe.</p>",

"<p>The fractional number of pixels in the subframe deemed " +
"free of image structure and used to estimate <i>Noise</i>.</p>",

"<p>The mean absolute deviation from the median star profile <i>full width at half " +
"maximum</i> (FWHM) estimate for the subframe in arcseconds or pixels.</p>",

"<p>The mean absolute deviation from the median star profile <i>eccentricity</i> " +
"estimate for the subframe.</p>",

"<p>The mean absolute deviation from the median <i>residual</i> of the star fitting " +
"process for the subframe.</p>",

"<p>The observation date and Coordinated Universal Time (UTC) of the " +
"subframe.</p>"
   ];

   this.medianRowTotal = function(total) {
      return "Subframe Median"
   }
   this.medianRowToolTip =
"<p>This row's columns contain the <i>median</i> of properties of across all " +
"subframes.</p>";

   this.meanDeviationRowTotal = function(total) {
      return "Subframe MeanDeviation"
   }
   this.meanDeviationRowToolTip =
"<p>This row's columns contain the <i>mean absolute deviation from the median</i> of " +
"properties of across all subframes.</p>";

   this.sortColumns = new Array(
      "Index",
      "Name",
      "Weight",
      "FWHM",
      "Eccentricity",
      "SNRWeight",
      "Median",
      "MeanDeviation",
      "Noise",
      "StarSupport",
      "StarResidual",
      "NoiseSupport",
      "FWHMMeanDev",
      "EccentricityMeanDev",
      "StarResidualMeanDev",
      "Date (UTC)"
   );
   this.sortColumnName = function(index) {
      if (index == 3) {
         return this.scaleUnit == 0 ? "FWHM (arcsec)" : "FWHM (pixel)";
      }
      if (index == 6) {
         return this.dataUnit == 0 ? "Median (e-)" : "Median (DN)";
      }
      if (index == 7) {
         return this.dataUnit == 0 ? "MeanDeviation (e-)" : "MeanDeviation (DN)";
      }
      if (index == 8) {
         return this.dataUnit == 0 ? "Noise (e-)" : "Noise (DN)";
      }
      if (index == 12) {
         return this.scaleUnit == 0 ? "FWHMMeanDev (arcsec)" : "FWHMMeanDev (pixel)";
      }
      return this.sortColumns[index];
   }

   this.sortOrderings = new Array(
      "Ascending",
      "Descending"
   );

   this.propertyPlotOrdinates = [
      "Weight",
      "FWHM",
      "Eccentricity",
      "SNRWeight",
      "Median",
      "MeanDeviation",
      "Noise",
      "StarSupport",
      "StarResidual",
      "NoiseSupport",
      "FWHMMeanDev",
      "EccentricityMeanDev",
      "StarResidualMeanDev"
   ];
   this.propertyPlotOrdinateName = function(index) {
      if (index == 1) {
         return this.scaleUnit == 0 ? "FWHM (arcsec)" : "FWHM (pixel)";
      }
      if (index == 4) {
         return this.dataUnit == 0 ? "Median (e-)" : "Median (DN)";
      }
      if (index == 5) {
         return this.dataUnit == 0 ? "MeanDeviation (e-)" : "MeanDeviation (DN)";
      }
      if (index == 6) {
         return this.dataUnit == 0 ? "Noise (e-)" : "Noise (DN)";
      }
      if (index == 10) {
         return this.scaleUnit == 0 ? "FWHMMeanDev (arcsec)" : "FWHMMeanDev (pixel)";
      }
      return this.propertyPlotOrdinates[index];
   }

   this.outputActions = new Array(
      "Copy",
      "Move",
      "None"
   );
   this.copyAction = 0;
   this.moveAction = 1;
   this.noneAction = 2;

   this.onErrors = new Array(
      "Continue",
      "Abort",
      "Ask User"
   );

   this.actualSubframeScale = function() {
      return this.scaleUnit == 0 ? this.subframeScale : 1.0;
   }
   this.actualCameraGain = function() {
      return this.dataUnit == 0 ? this.cameraGain : 1.0;
   }

   this.cacheSize = 2053;
   this.MD5MaxBytes = 1 * 1024 * 1024;

   this.minCropWidth = 256;
   this.minCropHeight = 256;

   this.minStarDescriptions = 300;
   this.minPartitionStarDescriptions = 30;
   this.maxPSFMAD = 0.1;

   this.currentPropertyPlotData = null;
   this.currentPropertyPlotDescripition = null;

   this.reset = function() {
      this.targetSubframeDescriptions = [];

      this.evaluationsDescriptions = [];

      for (var i = 0; i != this.parameterDescriptions.length; ++i) {
         var description = this.parameterDescriptions[i];
         var value = description.def;
         description.set(this, value);
       }
   };

   this.storeParameters = function() {
      Parameters.clear();

      this.targetSubframeDescriptionsLength = this.targetSubframeDescriptions.length;

      for (var i = 0; i != this.parameterDescriptions.length; ++i) {
         var description = this.parameterDescriptions[i];
         var value = description.get(this);
         description.storeParameters(description.name, value);
       }

      for (var i = 0; i != this.targetSubframeDescriptions.length; ++i) {
         var description = this.targetSubframeDescriptions[i];
         this.targetSubframeDescriptionChecked.storeParameters(
            this.targetSubframeDescriptionChecked.name + "_" + i, description.checked
         );
         this.targetSubframeDescriptionSelected.storeParameters(
            this.targetSubframeDescriptionSelected.name + "_" + i, description.selected
         );
         this.targetSubframeDescriptionFilePath.storeParameters(
            this.targetSubframeDescriptionFilePath.name + "_" + i, description.filePath
         );
      }
   };

   this.loadParameters = function() {
      this.targetSubframeDescriptions = [];

      this.evaluationsDescriptions = [];

      for (var i = 0; i != this.parameterDescriptions.length; ++i) {
         var description = this.parameterDescriptions[i];
         var value = description.loadParameters(description.name, description.def);
         description.set(this, value);
      }

      for (var i = 0; i != this.targetSubframeDescriptionsLength; ++i) {
         var checked = this.targetSubframeDescriptionChecked.loadParameters(
            this.targetSubframeDescriptionChecked.name + "_" + i
         );
         var selected = this.targetSubframeDescriptionSelected.loadParameters(
            this.targetSubframeDescriptionSelected.name + "_" + i
         );
         var filePath = this.targetSubframeDescriptionFilePath.loadParameters(
            this.targetSubframeDescriptionFilePath.name + "_" + i
         );
         if (filePath != "") {
            this.targetSubframeDescriptions.push(new targetSubframeDescription(
               checked,
               selected,
               filePath
            ));
         }
      }
   };

   this.storeSettings = function() {
      var prefix = TITLE + "." + VERSION + "_";

      this.targetSubframeDescriptionsLength = this.targetSubframeDescriptions.length;

      for (var i = 0; i != this.parameterDescriptions.length; ++i) {
         var description = this.parameterDescriptions[i];
         var value = description.get(this);
         description.storeSettings(prefix + description.name, value);
      }

      for (var i = 0; i != this.targetSubframeDescriptions.length; ++i) {
         var description = this.targetSubframeDescriptions[i];
         this.targetSubframeDescriptionChecked.storeSettings(
            prefix + this.targetSubframeDescriptionChecked.name + "_" + i,
            description.checked
         );
         this.targetSubframeDescriptionSelected.storeSettings(
            prefix + this.targetSubframeDescriptionSelected.name + "_" + i,
            description.selected
         );
         this.targetSubframeDescriptionFilePath.storeSettings(
            prefix + this.targetSubframeDescriptionFilePath.name + "_" + i,
            description.filePath
         );
      }
   };

   this.loadSettings = function() {
      var prefix = TITLE + "." + VERSION + "_";

      this.targetSubframeDescriptions = [];

      this.evaluationsDescriptions = [];

      for (var i = 0; i != this.parameterDescriptions.length; ++i) {
         var description = this.parameterDescriptions[i];
         var value = description.loadSettings(prefix + description.name, description.def);
         description.set(this, value);
      }

      for (var i = 0; i != this.targetSubframeDescriptionsLength; ++i) {
         var checked = this.targetSubframeDescriptionChecked.loadSettings(
            prefix + this.targetSubframeDescriptionChecked.name + "_" + i
         );
         var selected = this.targetSubframeDescriptionSelected.loadSettings(
            prefix + this.targetSubframeDescriptionSelected.name + "_" + i
         );
         var filePath = this.targetSubframeDescriptionFilePath.loadSettings(
            prefix + this.targetSubframeDescriptionFilePath.name + "_" + i
         );
         if (filePath != "") {
            this.targetSubframeDescriptions.push(new targetSubframeDescription(
               checked,
               selected,
               filePath
            ));
         }
      }
   };

   this.parameterDescriptions = [];
   this.insertParameterDescription = function(description) {
      this.parameterDescriptions.push(description);
   }
}
var parameters = new parametersPrototype();

// Parameter description
// name: parameter name
// def: default value
// set: function(parameters, value) {parameters.name = value;}
// get: function(parameters) {return parameters.name;}
// loadSettings: function(name, def) {load name from Settings with default;}
// storeSettings: function(name, value) {store name to Settings with value;}
// loadParameters: function(name, def) {load name from Parameters with default;}
// storeParameters: function(name, value) {store name to Parameters with value;}
function parameterDescription(
   name, def, get, set, loadSettings, storeSettings, loadParameters, storeParameters
) {
   this.name = name;
   this.def = def;
   this.get = get;
   this.set = set;
   this.loadSettings = loadSettings;
   this.storeSettings = storeSettings;
   this.loadParameters = loadParameters;
   this.storeParameters = storeParameters;
}

function defaultNullValue(value, def) {
   return value != null ? value : def;
}

function defaultNullMinMaxValue(value, min, max, def) {
   return value != null && !isNaN(value) && value >= min && value <= max ? value : def;
}

function loadSettingsDataTypeString() {
   return function(name, def) {
      return defaultNullValue(Settings.read(name, DataType_String), def);
   };
}

function storeSettingsDataTypeString() {
   return function(name, value) {
      Settings.write(name, DataType_String, value);
   };
}

function loadParametersDataTypeString() {
   return function(name, def) {
      return Parameters.has(name) ?
         defaultNullValue(Parameters.getString(name), def) :
         def;
   };
}

function storeParametersDataTypeString() {
   return function(name, value) {
      Parameters.set(name, value);
   };
}

function loadSettingsDataTypeBoolean() {
   return function(name, def) {
      return defaultNullValue(Settings.read(name, DataType_Boolean), def);
   };
}

function storeSettingsDataTypeBoolean() {
   return function(name, value) {
      Settings.write(name, DataType_Boolean, value);
   };
}

function loadParametersDataTypeBoolean() {
   return function(name, def) {
      return Parameters.has(name) ?
         defaultNullValue(Parameters.getBoolean(name), def) :
         def;
   };
}

function storeParametersDataTypeBoolean() {
   return function(name, value) {
      Parameters.set(name, value);
   };
}

function loadSettingsDataTypeInt32(min, max) {
   return function(name, def) {
      return defaultNullMinMaxValue(Settings.read(name, DataType_Int32), min, max, def);
   };
}

function storeSettingsDataTypeInt32() {
   return function(name, value) {
      Settings.write(name, DataType_Int32, value);
   };
}

function loadParametersDataTypeInt32(min, max) {
   return function(name, def) {
      return Parameters.has(name) ?
         defaultNullMinMaxValue(Parameters.getInteger(name), min, max, def) :
         def;
   };
}

function storeParametersDataTypeInt32() {
   return function(name, value) {
      Parameters.set(name, value);
   };
}

function loadSettingsDataTypeReal32(min, max) {
   return function(name, def) {
      return defaultNullMinMaxValue(Settings.read(name, DataType_Real32), min, max, def);
   };
}

function storeSettingsDataTypeReal32() {
   return function(name, value) {
      Settings.write(name, DataType_Real32, value);
   };
}

function loadParametersDataTypeReal32(min, max) {
   return function(name, def) {
      return Parameters.has(name) ?
         defaultNullMinMaxValue(Parameters.getReal(name), min, max, def) :
         def;
   };
}

function storeParametersDataTypeReal32() {
   return function(name, value) {
      Parameters.set(name, value);
   };
}

parameters.targetSubframeDescriptionChecked = new parameterDescription(
   "targetSubframeDescriptionChecked",
   true,
   function(parameters) {
      return true;
   },
   function(parameters, value) {
   },
   loadSettingsDataTypeBoolean(),
   storeSettingsDataTypeBoolean(),
   loadParametersDataTypeBoolean(),
   storeParametersDataTypeBoolean()
);

parameters.targetSubframeDescriptionSelected = new parameterDescription(
   "targetSubframeDescriptionSelected",
   false,
   function(parameters) {
      return false;
   },
   function(parameters, value) {
   },
   loadSettingsDataTypeBoolean(),
   storeSettingsDataTypeBoolean(),
   loadParametersDataTypeBoolean(),
   storeParametersDataTypeBoolean()
);

parameters.targetSubframeDescriptionFilePath = new parameterDescription(
   "targetSubframeDescriptionFilePath",
   "",
   function(parameters) {
      return "";
   },
   function(parameters, value) {
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
);

parameters.insertParameterDescription(new parameterDescription(
   "targetSubframeDescriptionsLength",
   0,
   function(parameters) {
      return parameters.targetSubframeDescriptionsLength;
   },
   function(parameters, value) {
      parameters.targetSubframeDescriptionsLength = value;
   },
   loadSettingsDataTypeInt32(0, 1024),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 1024),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "targetSubframesFullPaths",
   false,
   function(parameters) {
      return parameters.targetSubframesFullPaths;
   },
   function(parameters, value) {
      parameters.targetSubframesFullPaths = value;
   },
   loadSettingsDataTypeBoolean(),
   storeSettingsDataTypeBoolean(),
   loadParametersDataTypeBoolean(),
   storeParametersDataTypeBoolean()
));

parameters.insertParameterDescription(new parameterDescription(
   "useFileCache",
   false,
   function(parameters) {
      return false; // parameters.useFileCache;
   },
   function(parameters, value) {
      parameters.useFileCache = false; // value;
   },
   loadSettingsDataTypeBoolean(),
   storeSettingsDataTypeBoolean(),
   loadParametersDataTypeBoolean(),
   storeParametersDataTypeBoolean()
));

parameters.insertParameterDescription(new parameterDescription(
   "subframeScale",
   1.0,
   function(parameters) {
      return parameters.subframeScale;
   },
   function(parameters, value) {
      parameters.subframeScale = value;
   },
   loadSettingsDataTypeReal32(0.01, 100.0),
   storeSettingsDataTypeReal32(),
   loadParametersDataTypeReal32(0.01, 100.0),
   storeParametersDataTypeReal32()
));

parameters.insertParameterDescription(new parameterDescription(
   "cameraGain",
   1.0,
   function(parameters) {
      return parameters.cameraGain;
   },
   function(parameters, value) {
      parameters.cameraGain = value;
   },
   loadSettingsDataTypeReal32(0.01, 100.0),
   storeSettingsDataTypeReal32(),
   loadParametersDataTypeReal32(0.01, 100.0),
   storeParametersDataTypeReal32()
));

parameters.insertParameterDescription(new parameterDescription(
   "cameraResolution",
   4,
   function(parameters) {
      return parameters.cameraResolution;
   },
   function(parameters, value) {
      parameters.cameraResolution = value;
   },
   loadSettingsDataTypeInt32(0, parameters.cameraResolutions.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.cameraResolutions.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "subframeRegionLeft",
   0,
   function(parameters) {
      return parameters.subframeRegionLeft;
   },
   function(parameters, value) {
      parameters.subframeRegionLeft = value;
   },
   loadSettingsDataTypeInt32(0, 1024 * 1024 * 1024 - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 1024 * 1024 * 1024 - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "subframeRegionTop",
   0,
   function(parameters) {
      return parameters.subframeRegionTop;
   },
   function(parameters, value) {
      parameters.subframeRegionTop = value;
   },
   loadSettingsDataTypeInt32(0, 1024 * 1024 * 1024 - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 1024 * 1024 * 1024 - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "subframeRegionWidth",
   0,
   function(parameters) {
      return parameters.subframeRegionWidth;
   },
   function(parameters, value) {
      parameters.subframeRegionWidth = value;
   },
   loadSettingsDataTypeInt32(0, 1024 * 1024 * 1024 - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 1024 * 1024 * 1024 - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "subframeRegionHeight",
   0,
   function(parameters) {
      return parameters.subframeRegionHeight;
   },
   function(parameters, value) {
      parameters.subframeRegionHeight = value;
   },
   loadSettingsDataTypeInt32(0, 1024 * 1024 * 1024 - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 1024 * 1024 * 1024 - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "pedestal",
   0,
   function(parameters) {
      return parameters.pedestal;
   },
   function(parameters, value) {
      parameters.pedestal = value;
   },
   loadSettingsDataTypeInt32(0, 65535),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 65535),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "siteLocalMidnight",
   24,
   function(parameters) {
      return parameters.siteLocalMidnight;
   },
   function(parameters, value) {
      parameters.siteLocalMidnight = value;
   },
   loadSettingsDataTypeInt32(0, 24),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 24),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "scaleUnit",
   1,
   function(parameters) {
      return parameters.scaleUnit;
   },
   function(parameters, value) {
      parameters.scaleUnit = value;
   },
   loadSettingsDataTypeInt32(0, parameters.scaleUnits.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.scaleUnits.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "dataUnit",
   1,
   function(parameters) {
      return parameters.dataUnit;
   },
   function(parameters, value) {
      parameters.dataUnit = value;
   },
   loadSettingsDataTypeInt32(0, parameters.dataUnits.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.dataUnits.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "starDetectionLayers",
   5,
   function(parameters) {
      return parameters.starDetectionLayers;
   },
   function(parameters, value) {
      parameters.starDetectionLayers = value;
   },
   loadSettingsDataTypeInt32(1, 8),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(1, 8),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "noiseReductionLayers",
   1,
   function(parameters) {
      return parameters.noiseReductionLayers;
   },
   function(parameters, value) {
      parameters.noiseReductionLayers = value;
   },
   loadSettingsDataTypeInt32(0, 4),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 4),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "hotPixelFilterRadius",
   1,
   function(parameters) {
      return parameters.hotPixelFilterRadius;
   },
   function(parameters, value) {
      parameters.hotPixelFilterRadius = value;
   },
   loadSettingsDataTypeInt32(0, 2),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, 2),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "starDetectionSensitivity",
   0.1,
   function(parameters) {
      return parameters.starDetectionSensitivity;
   },
   function(parameters, value) {
      parameters.starDetectionSensitivity = value;
   },
   loadSettingsDataTypeReal32(0.001, 1000.0),
   storeSettingsDataTypeReal32(),
   loadParametersDataTypeReal32(0.001, 1000.0),
   storeParametersDataTypeReal32()
));

parameters.insertParameterDescription(new parameterDescription(
   "starPeakResponse",
   0.8,
   function(parameters) {
      return parameters.starPeakResponse;
   },
   function(parameters, value) {
      parameters.starPeakResponse = value;
   },
   loadSettingsDataTypeReal32(0.0, 1.0),
   storeSettingsDataTypeReal32(),
   loadParametersDataTypeReal32(0.0, 1.0),
   storeParametersDataTypeReal32()
));

parameters.insertParameterDescription(new parameterDescription(
   "maximumStarDistortion",
   0.5,
   function(parameters) {
      return parameters.maximumStarDistortion;
   },
   function(parameters, value) {
      parameters.maximumStarDistortion = value;
   },
   loadSettingsDataTypeReal32(0.0, 1.0),
   storeSettingsDataTypeReal32(),
   loadParametersDataTypeReal32(0.0, 1.0),
   storeParametersDataTypeReal32()
));

parameters.insertParameterDescription(new parameterDescription(
   "upperLimit",
   1.0,
   function(parameters) {
      return parameters.upperLimit;
   },
   function(parameters, value) {
      parameters.upperLimit = value;
   },
   loadSettingsDataTypeReal32(0.0, 1.0),
   storeSettingsDataTypeReal32(),
   loadParametersDataTypeReal32(0.0, 1.0),
   storeParametersDataTypeReal32()
));

parameters.insertParameterDescription(new parameterDescription(
   "modelFunction",
   4,
   function(parameters) {
      return parameters.modelFunction;
   },
   function(parameters, value) {
      parameters.modelFunction = value;
   },
   loadSettingsDataTypeInt32(0, parameters.modelFunctions.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.modelFunctions.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "circularModel",
   false,
   function(parameters) {
      return parameters.circularModel;
   },
   function(parameters, value) {
      parameters.circularModel = value;
   },
   loadSettingsDataTypeBoolean(),
   storeSettingsDataTypeBoolean(),
   loadParametersDataTypeBoolean(),
   storeParametersDataTypeBoolean()
));

parameters.insertParameterDescription(new parameterDescription(
   "applyHotPixelFilterToDetectionImage",
   false,
   function(parameters) {
      return parameters.applyHotPixelFilterToDetectionImage;
   },
   function(parameters, value) {
      parameters.applyHotPixelFilterToDetectionImage = value;
   },
   loadSettingsDataTypeBoolean(),
   storeSettingsDataTypeBoolean(),
   loadParametersDataTypeBoolean(),
   storeParametersDataTypeBoolean()
));

parameters.insertParameterDescription(new parameterDescription(
   "selectorExpression",
   "",
   function(parameters) {
      return parameters.selectorExpression;
   },
   function(parameters, value) {
      parameters.selectorExpression = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "weightingExpression",
   "SNRWeight",
   function(parameters) {
      return parameters.weightingExpression;
   },
   function(parameters, value) {
      parameters.weightingExpression = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "sortColumn",
   0,
   function(parameters) {
      return parameters.sortColumn;
   },
   function(parameters, value) {
      parameters.sortColumn = value;
   },
   loadSettingsDataTypeInt32(0, parameters.sortColumns.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.sortColumns.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "sortOrdering",
   0,
   function(parameters) {
      return parameters.sortOrdering;
   },
   function(parameters, value) {
      parameters.sortOrdering = value;
   },
   loadSettingsDataTypeInt32(0, parameters.sortOrderings.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.sortOrderings.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "propertyPlotOrdinate",
   1,
   function(parameters) {
      return parameters.propertyPlotOrdinate;
   },
   function(parameters, value) {
      parameters.propertyPlotOrdinate = value;
   },
   loadSettingsDataTypeInt32(0, parameters.propertyPlotOrdinates.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.propertyPlotOrdinates.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "approvedAction",
   0,
   function(parameters) {
      return parameters.approvedAction;
   },
   function(parameters, value) {
      parameters.approvedAction = value;
   },
   loadSettingsDataTypeInt32(0, parameters.outputActions.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.outputActions.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "approvedDirectory",
   "",
   function(parameters) {
      return parameters.approvedDirectory;
   },
   function(parameters, value) {
      parameters.approvedDirectory = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "approvedPostfix",
   "_a",
   function(parameters) {
      return parameters.approvedPostfix;
   },
   function(parameters, value) {
      parameters.approvedPostfix = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "rejectedAction",
   2,
   function(parameters) {
      return parameters.rejectedAction;
   },
   function(parameters, value) {
      parameters.rejectedAction = value;
   },
   loadSettingsDataTypeInt32(0, parameters.outputActions.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.outputActions.length - 1),
   storeParametersDataTypeInt32()
));

parameters.insertParameterDescription(new parameterDescription(
   "rejectedDirectory",
   "",
   function(parameters) {
      return parameters.rejectedDirectory;
   },
   function(parameters, value) {
      parameters.rejectedDirectory = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "rejectedPostfix",
   "_x",
   function(parameters) {
      return parameters.rejectedPostfix;
   },
   function(parameters, value) {
      parameters.rejectedPostfix = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "starMapDirectory",
   "",
   function(parameters) {
      return parameters.starMapDirectory;
   },
   function(parameters, value) {
      parameters.starMapDirectory = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "weightKeyword",
   "",
   function(parameters) {
      return parameters.weightKeyword;
   },
   function(parameters, value) {
      parameters.weightKeyword = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "starMapPostfix",
   "_s_m",
   function(parameters) {
      return parameters.starMapPostfix;
   },
   function(parameters, value) {
      parameters.starMapPostfix = value;
   },
   loadSettingsDataTypeString(),
   storeSettingsDataTypeString(),
   loadParametersDataTypeString(),
   storeParametersDataTypeString()
));

parameters.insertParameterDescription(new parameterDescription(
   "overwriteExistingFiles",
   false,
   function(parameters) {
      return parameters.overwriteExistingFiles;
   },
   function(parameters, value) {
      parameters.overwriteExistingFiles = value;
   },
   loadSettingsDataTypeBoolean(),
   storeSettingsDataTypeBoolean(),
   loadParametersDataTypeBoolean(),
   storeParametersDataTypeBoolean()
));

parameters.insertParameterDescription(new parameterDescription(
   "onError",
   2,
   function(parameters) {
      return parameters.onError;
   },
   function(parameters, value) {
      parameters.onError = value;
   },
   loadSettingsDataTypeInt32(0, parameters.onErrors.length - 1),
   storeSettingsDataTypeInt32(),
   loadParametersDataTypeInt32(0, parameters.onErrors.length - 1),
   storeParametersDataTypeInt32()
));

parameters.reset();
// end section
//
// Copyright (C) 2012-2016 Mike Schuster. All Rights Reserved.
// Copyright (C) 2003-2016 Pleiades Astrophoto S.L. All Rights Reserved.

function PreviewDialog(mainDialog)
{
   // establish dialog base

   this.__base__ = Dialog;
   this.__base__();

	this.mainDialog = mainDialog;
   this.table  = mainDialog.table;
   this.header = mainDialog.header;
   this.types  = mainDialog.types;
   this.fullNames = mainDialog.fullNames;
   this.sortColumn = 0;
   this.sortMode   = true; // ASC

   this.lblPrecision  = new Label(this);
   this.lblPrecision.text = "Precision:";

   this.numPrecision = new SpinBox( this );
   this.numPrecision.minValue = 0;
   this.numPrecision.maxValue = 15;
   this.numPrecision.value = this.mainDialog.precision;
   this.numPrecision.toolTip = "<p>Set decimal numbers.</p>";
   this.numPrecision.onValueUpdated = function( value )
   {
      this.parent.mainDialog.precision = value;
   };

   this.btnRefesh = new PushButton(this);
   this.btnRefesh.text = "Refresh";
   this.btnRefesh.onClick = function(x)
   {
      fillGrid(this.parent);
   }

   this.topSizer = new HorizontalSizer(this);
   this.topSizer.margin = 4;
   this.topSizer.addSpacing( 8 );
   this.topSizer.add(this.lblPrecision);
   this.topSizer.addSpacing( 8 );
   this.topSizer.add(this.numPrecision);
   this.topSizer.addSpacing( 8 );
   this.topSizer.add(this.btnRefesh);
   this.topSizer.addStretch();

   this.grid = new TreeBox(this);
   with (this.grid)
   {
      alternateRowColor = true;
      rootDecoration  = false;
      headerSorting  = true;
      setMinSize(800, 600);

      fillGrid(this);

		onNodeDoubleClicked = function( item, columnIndex )
		{
			if (columnIndex >-1) // == 0)
			{
				var filename = item.text(0);
				for (var i in this.dialog.mainDialog.filesCollection)
				{
					if (this.dialog.mainDialog.filesCollection[i].compare(filename))
					{
						viewFileObject(this.dialog.mainDialog.filesCollection[i]);
						break;
					}
				}
			}
		}
	}

   this.sizer= new VerticalSizer();
   this.sizer.add(this.topSizer);
   this.sizer.add(this.grid);
   this.windowTitle = "Fits Data Summary";
   this.adjustToContents();
	this.userResizable = true;

   function fillGrid(dialog)
   {
      if (dialog.table.length > 0)
      {
         dialog.grid.clear();
         var cols = dialog.header.length;
         for (var i = 0; i < cols; i++)
         {
				dialog.grid.setHeaderText(i, dialog.header[i]);
         }

         for (var i = 0; i < dialog.table.length; i++)
         {
            var row  = dialog.table[i];
            var node = new TreeBoxNode(dialog.grid);
            node.checkable = false;

            for (var j = 0; j <  cols; j++)
            {
               var v = row[j];

               if (j == 0 && !dialog.fullNames) v = File.extractNameAndExtension(v);

               var t = dialog.types[j];

               var s = "";
               if (t == 0)
                  s = v;
               else if (t == 1)
                  s = v.toString();
               else if (t == 2)
                  s = v.toFixed(dialog.mainDialog.precision);
               else
                  s = v.toString();
               node.setText(j, s);
            }
         }
         for (var i = 0; i < cols; i++)
            dialog.grid.adjustColumnWidthToContents(i);
      }
   }
}

function ViewDialog(fo)
{
   // establish dialog base

   this.__base__ = Dialog;
   this.__base__();

 	this.view = fo.window.currentView;

	//
	// FITS files must be stretched
	//

	ApplyAutoSTF( this.view,
						DEFAULT_AUTOSTRETCH_SCLIP,
						DEFAULT_AUTOSTRETCH_TBGND,
						DEFAULT_AUTOSTRETCH_CLINK );

	var stf = this.view.stf;

	var H = [[  0, 0.0, 1.0, 0, 1.0],
				[  0, 0.5, 1.0, 0, 1.0],
				[  0, 0.5, 1.0, 0, 1.0],
				[  0, 0.5, 1.0, 0, 1.0],
				[  0, 0.5, 1.0, 0, 1.0]];;

	if (this.view.image.isColor)
	{
		for (var c = 0; c < 3; c++)
		{
			H[c][0] = stf[c][1];
			H[c][1] = stf[c][0];
		}
	}
	else
	{
		H[3][0] = stf[0][1];
		H[3][1] = stf[0][0];
	}

	var STF = new ScreenTransferFunction;

	this.view.stf =  [ // c0, c1, m, r0, r1
	[0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
	[0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
	[0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
	[0.00000, 1.00000, 0.50000, 0.00000, 1.00000]
	];

	STF.executeOn(this.view)

	var HT = new HistogramTransformation;
	HT.H = H;
	HT.executeOn(this.view);

   this.bitmap = GetWindowBmp(this.view.window);

	this.previewControl = new PreviewControl(this);
	this.previewControl.setMinSize(fo.window.viewportWidth + 20,
      fo.window.viewportHeight +
      this.previewControl.coords_Frame.height +
      this.previewControl.zoomIn_Button.height + 20);
	this.previewControl.SetImage(this.bitmap, this.bitmap);
   this.previewControl.UpdateZoom(fo.window.zoomFactor);

   this.sizer= new Sizer();
	this.sizer.add(this.previewControl);
   this.windowTitle = "View " + fo.shortName;
   this.adjustToContents();
	this.userResizable = true;
}

function viewFileObject(fo)
{
	Console.writeln("View File "+fo.name);
	var dialog = new ViewDialog(fo);
   dialog.execute();
   fo.window.currentView.setPropertyValue( "dispose", true )
}

// ********************************************************************************************
// Copyright (C) 2013, Andres del Pozo
//
// ********************************************************************************************
function PreviewControl(parent)
{
	this.__base__ = Frame;
  	this.__base__(parent);

   this.SetImage = function(image, metadata)
   {
      this.image = image;
      this.metadata = metadata;
      this.scaledImage = null;
      this.SetZoomOutLimit();
      this.UpdateZoom(-100);
      //this.stf_Button.visible = true;
   }

   this.UpdateZoom = function (newZoom, refPoint)
   {
      newZoom = Math.max(this.zoomOutLimit, Math.min(2, newZoom));
      if (newZoom == this.zoom && this.scaledImage)
         return;

      if(refPoint==null)
         refPoint=new Point(this.scrollbox.viewport.width/2, this.scrollbox.viewport.height/2);
      var imgx=null;
      if(this.scrollbox.maxHorizontalScrollPosition>0)
         imgx=(refPoint.x+this.scrollbox.horizontalScrollPosition)/this.scale;
      var imgy=null;
      if(this.scrollbox.maxVerticalScrollPosition>0)
         imgy=(refPoint.y+this.scrollbox.verticalScrollPosition)/this.scale;

      this.zoom = newZoom;
      this.scaledImage = null;
      gc(true);
      if (this.zoom > 0)
      {
         this.scale = this.zoom;
         this.zoomVal_Label.text = format("%d:1", this.zoom);
      }
      else
      {
         this.scale = 1 / (-this.zoom + 2);
         this.zoomVal_Label.text = format("1:%d", -this.zoom + 2);
      }
      if (this.image)
         this.scaledImage = this.image.scaled(this.scale);
      else
         this.scaledImage = {width:this.metadata.width * this.scale, height:this.metadata.height * this.scale};
      this.scrollbox.maxHorizontalScrollPosition = Math.max(0, this.scaledImage.width - this.scrollbox.viewport.width);
      this.scrollbox.maxVerticalScrollPosition = Math.max(0, this.scaledImage.height - this.scrollbox.viewport.height);

      if(this.scrollbox.maxHorizontalScrollPosition>0 && imgx!=null)
         this.scrollbox.horizontalScrollPosition = (imgx*this.scale)-refPoint.x;
      if(this.scrollbox.maxVerticalScrollPosition>0 && imgy!=null)
         this.scrollbox.verticalScrollPosition = (imgy*this.scale)-refPoint.y;

      this.scrollbox.viewport.update();
   }

	this.zoomIn_Button = new ToolButton( this );
   this.zoomIn_Button.icon = this.scaledResource( ":/icons/zoom-in.png" );
   this.zoomIn_Button.setScaledFixedSize( 20, 20 );
   this.zoomIn_Button.toolTip = "Zoom in";
   this.zoomIn_Button.onMousePress = function()
   {
      this.parent.UpdateZoom(this.parent.zoom+1);
   };

   this.zoomOut_Button = new ToolButton( this );
   this.zoomOut_Button.icon = this.scaledResource( ":/icons/zoom-out.png" );
   this.zoomOut_Button.setScaledFixedSize( 20, 20 );
   this.zoomOut_Button.toolTip = "Zoom in";
   this.zoomOut_Button.onMousePress = function()
   {
      this.parent.UpdateZoom(this.parent.zoom-1);
   };


   this.zoom11_Button = new ToolButton( this );
   this.zoom11_Button.icon = this.scaledResource( ":/icons/zoom-1-1.png" );
   this.zoom11_Button.setScaledFixedSize( 20, 20 );
   this.zoom11_Button.toolTip = "Zoom 1:1";
   this.zoom11_Button.onMousePress = function()
   {
      this.parent.UpdateZoom(1);
   };

/*
 * this.stf_Button =  new ToolButton( this );
   this.stf_Button.text = "STF";
   this.stf_Button.visible = false;
   this.stf_Button.toolTip = "ScreenTransferFunction";
   this.stf_Button.onMousePress = function()
   {
      var preview = this.parent;
      preview.onCustomSTF.call(this);
   }
*/

   this.buttons_Sizer = new HorizontalSizer;
   this.buttons_Sizer.add( this.zoomIn_Button );
   this.buttons_Sizer.add( this.zoomOut_Button );
   this.buttons_Sizer.add( this.zoom11_Button );
   this.buttons_Sizer.addStretch();
 	//this.buttons_Sizer.add( this.stf_Button );

   this.setScaledMinSize(300,200);
   this.zoom = 1;
   this.scale = 1;
   this.zoomOutLimit = -5;
   this.scrollbox = new ScrollBox(this);
   this.scrollbox.autoScroll = true;
   this.scrollbox.tracking = true;
   this.scrollbox.cursor = new Cursor(StdCursor_Arrow);

   this.scroll_Sizer = new HorizontalSizer;
   this.scroll_Sizer .add( this.scrollbox );

   this.SetZoomOutLimit = function()
   {
      var scaleX = Math.ceil(this.metadata.width/this.scrollbox.viewport.width);
      var scaleY = Math.ceil(this.metadata.height/this.scrollbox.viewport.height);
      var scale = Math.max(scaleX,scaleY);
      this.zoomOutLimit = -scale+2;
   }

   this.scrollbox.onHorizontalScrollPosUpdated = function (newPos)
   {
      this.viewport.update();
   }
   this.scrollbox.onVerticalScrollPosUpdated = function (newPos)
   {
      this.viewport.update();
   }

   this.forceRedraw = function()
   {
      this.scrollbox.viewport.update();
   };

   this.scrollbox.viewport.onMouseWheel = function (x, y, delta, buttonState, modifiers)
   {
      var preview = this.parent.parent;
      preview.UpdateZoom(preview.zoom + (delta > 0 ? -1 : 1), new Point(x,y));
   }

   this.scrollbox.viewport.onMousePress = function ( x, y, button, buttonState, modifiers )
   {
      var preview = this.parent.parent;
      var p =  preview.transform(x, y, preview);
      if(preview.onCustomMouseDown)
      {
         preview.onCustomMouseDown.call(this, p.x, p.y, button, buttonState, modifiers )
      }
   }

   this.scrollbox.viewport.onMouseMove = function ( x, y, buttonState, modifiers )
   {
      var preview = this.parent.parent;
      var p =  preview.transform(x, y, preview);
      preview.Xval_Label.text = p.x.toString();
      preview.Yval_Label.text = p.y.toString();

      if(preview.onCustomMouseMove)
      {
         preview.onCustomMouseMove.call(this, p.x, p.y, buttonState, modifiers )
      }
   }

   this.scrollbox.viewport.onMouseRelease = function (x, y, button, buttonState, modifiers)
   {
      var preview = this.parent.parent;

      var p =  preview.transform(x, y, preview);
      if(preview.onCustomMouseUp)
      {
         preview.onCustomMouseUp.call(this, p.x, p.y, button, buttonState, modifiers )
      }
   }

   this.scrollbox.viewport.onResize = function (wNew, hNew, wOld, hOld)
   {
      var preview = this.parent.parent;
      if(preview.metadata && preview.scaledImage)
      {
         this.parent.maxHorizontalScrollPosition = Math.max(0, preview.scaledImage.width - wNew);
         this.parent.maxVerticalScrollPosition = Math.max(0, preview.scaledImage.height - hNew);
         preview.SetZoomOutLimit();
         preview.UpdateZoom(preview.zoom);
      }
      this.update();
   }

   this.scrollbox.viewport.onPaint = function (x0, y0, x1, y1)
   {
      var preview = this.parent.parent;
      var graphics = new VectorGraphics(this);

      graphics.fillRect(x0,y0, x1, y1, new Brush(0xff202020));
      var offsetX = this.parent.maxHorizontalScrollPosition>0 ? -this.parent.horizontalScrollPosition : (this.width-preview.scaledImage.width)/2;
      var offsetY = this.parent.maxVerticalScrollPosition>0 ? -this.parent.verticalScrollPosition: (this.height-preview.scaledImage.height)/2;
      graphics.translateTransformation(offsetX, offsetY);
      if(preview.image)
         graphics.drawBitmap(0, 0, preview.scaledImage);
      else
         graphics.fillRect(0, 0, preview.scaledImage.width, preview.scaledImage.height, new Brush(0xff000000));

      graphics.pen = new Pen(0xffffffff,0);
      graphics.drawRect(-1, -1, preview.scaledImage.width + 1, preview.scaledImage.height + 1);

      if(preview.onCustomPaint)
      {
         graphics.antialiasing = true;
         graphics.scaleTransformation(preview.scale,preview.scale);
         preview.onCustomPaint.call(this, graphics, x0, y0, x1, y1);
      }
      graphics.end();
   }

   this.transform = function(x, y, preview)
   {
      var scrollbox = preview.scrollbox;
      var ox = 0;
      var oy = 0;
      ox = scrollbox.maxHorizontalScrollPosition>0 ? -scrollbox.horizontalScrollPosition : (scrollbox.viewport.width-preview.scaledImage.width)/2;
      oy = scrollbox.maxVerticalScrollPosition>0 ? -scrollbox.verticalScrollPosition: (scrollbox.viewport.height-preview.scaledImage.height)/2;
      var coordPx = new Point((x - ox) / preview.scale, (y - oy) / preview.scale);
      return new Point(coordPx.x, coordPx.y);
   }

   this.center = function()
   {
      var preview = this;
      var scrollbox = preview.scrollbox;
      var x = scrollbox.viewport.width / 2;
      var y = scrollbox.viewport.height / 2;
      var p =  this.transform(x, y, preview);
      return p;
   }

   this.zoomLabel_Label =new Label(this);
   this.zoomLabel_Label.text = "Zoom:";
   this.zoomVal_Label =new Label(this);
   this.zoomVal_Label.text = "1:1";

   this.Xlabel_Label = new Label(this);
   this.Xlabel_Label .text = "X:";
   this.Xval_Label = new Label(this);
   this.Xval_Label.text = "---";
   this.Ylabel_Label = new Label(this);
   this.Ylabel_Label.text = "Y:";
   this.Yval_Label = new Label(this);
   this.Yval_Label.text = "---";

   this.coords_Frame = new Frame(this);
   this.coords_Frame.backgroundColor = 0xffffffff;
   this.coords_Frame.sizer = new HorizontalSizer;
   this.coords_Frame.sizer.margin = 2;
   this.coords_Frame.sizer.spacing = 4;
   this.coords_Frame.sizer.add(this.zoomLabel_Label);
   this.coords_Frame.sizer.add(this.zoomVal_Label);
   this.coords_Frame.sizer.addSpacing(6);
   this.coords_Frame.sizer.add(this.Xlabel_Label);
   this.coords_Frame.sizer.add(this.Xval_Label);
   this.coords_Frame.sizer.addSpacing(6);
   this.coords_Frame.sizer.add(this.Ylabel_Label);
   this.coords_Frame.sizer.add(this.Yval_Label);

   this.coords_Frame.sizer.addStretch();

   this.sizer = new VerticalSizer;
   this.sizer.add(this.buttons_Sizer);
   this.sizer.add(this.scroll_Sizer);
   this.sizer.add(this.coords_Frame);
}

PreviewControl.prototype = new Frame;

// ************************************ PreviewControl definition end
/*
 * STF Auto Stretch routine
 */
function ApplyAutoSTF( view, shadowsClipping, targetBackground, rgbLinked )
{
   var stf = new ScreenTransferFunction;

   var n = view.image.isColor ? 3 : 1;

   var median = view.computeOrFetchProperty( "Median" );

   var mad = view.computeOrFetchProperty( "MAD" );
   mad.mul( 1.4826 ); // coherent with a normal distribution

   if ( rgbLinked )
   {
      /*
       * Try to find how many channels look as channels of an inverted image.
       * We know a channel has been inverted because the main histogram peak is
       * located over the right-hand half of the histogram. Seems simplistic
       * but this is consistent with astronomical images.
       */
      var invertedChannels = 0;
      for ( var c = 0; c < n; ++c )
         if ( median.at( c ) > 0.5 )
            ++invertedChannels;

      if ( invertedChannels < n )
      {
         /*
          * Noninverted image
          */
         var c0 = 0, m = 0;
         for ( var c = 0; c < n; ++c )
         {
            if ( 1 + mad.at( c ) != 1 )
               c0 += median.at( c ) + shadowsClipping * mad.at( c );
            m  += median.at( c );
         }
         c0 = Math.range( c0/n, 0.0, 1.0 );
         m = Math.mtf( targetBackground, m/n - c0 );

         stf.STF = [ // c0, c1, m, r0, r1
                     [c0, 1, m, 0, 1],
                     [c0, 1, m, 0, 1],
                     [c0, 1, m, 0, 1],
                     [0, 1, 0.5, 0, 1] ];
      }
      else
      {
         /*
          * Inverted image
          */
         var c1 = 0, m = 0;
         for ( var c = 0; c < n; ++c )
         {
            m  += median.at( c );
            if ( 1 + mad.at( c ) != 1 )
               c1 += median.at( c ) - shadowsClipping * mad.at( c );
            else
               c1 += 1;
         }
         c1 = Math.range( c1/n, 0.0, 1.0 );
         m = Math.mtf( c1 - m/n, targetBackground );

         stf.STF = [ // c0, c1, m, r0, r1
                     [0, c1, m, 0, 1],
                     [0, c1, m, 0, 1],
                     [0, c1, m, 0, 1],
                     [0, 1, 0.5, 0, 1] ];
      }
   }
   else
   {
      /*
       * Unlinked RGB channnels: Compute automatic stretch functions for
       * individual RGB channels separately.
       */
      var A = [ // c0, c1, m, r0, r1
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1] ];

      for ( var c = 0; c < n; ++c )
      {
         if ( median.at( c ) < 0.5 )
         {
            /*
             * Noninverted channel
             */
            var c0 = (1 + mad.at( c ) != 1) ? Math.range( median.at( c ) + shadowsClipping * mad.at( c ), 0.0, 1.0 ) : 0.0;
            var m  = Math.mtf( targetBackground, median.at( c ) - c0 );
            A[c] = [c0, 1, m, 0, 1];
         }
         else
         {
            /*
             * Inverted channel
             */
            var c1 = (1 + mad.at( c ) != 1) ? Math.range( median.at( c ) - shadowsClipping * mad.at( c ), 0.0, 1.0 ) : 1.0;
            var m  = Math.mtf( c1 - median.at( c ), targetBackground );
            A[c] = [0, c1, m, 0, 1];
         }
      }

      stf.STF = A;
   }
   stf.executeOn( view );

//   console.writeln( "<end><cbr/><br/>" );
}


function GetWindowBmp(window)
{
   var imageOrg = window.mainView.image;
   var tmpW = null;
   try
   {
      tmpW = new ImageWindow(imageOrg.width, imageOrg.height, imageOrg.numberOfChannels,
         window.bitsPerSample, window.isFloatSample, imageOrg.isColor, "Aux");
      tmpW.mainView.beginProcess(UndoFlag_NoSwapFile);
      tmpW.mainView.image.apply(imageOrg);
     // ApplySTF(tmpW.mainView, window.mainView.stf);
      tmpW.mainView.endProcess();
      var bmp = new Bitmap(imageOrg.width, imageOrg.height);
      bmp.assign(tmpW.mainView.image.render());
      return bmp;
   } finally
   {
      tmpW.forceClose();
   }
}

function getAllMainViews()
{
   var mainViews = [];
   var images = ImageWindow.windows;
   for ( var i in images )
   {
      if (images[i].mainView.isMainView) mainViews.push(images[i].mainView);
   }
   return mainViews;
}


FDVDialog.prototype = new Dialog;
PreviewDialog.prototype = new Dialog;
ViewDialog.prototype = new Dialog;

function main()
{
   Console.abortEnabled = true;
   Console.show();
   //Console.hide();
   var dialog = new FDVDialog();
   dialog.execute();

   var views = getAllMainViews();
   for (var i = 0; i < views.length; i++)
   {
      if (views[i].propertyValue("dispose" ))
      {
         views[i].window.forceClose();
      }
   }
}

main();

